"""
Data model class hierarchy
"""

import copy
import datetime
import os
from pathlib import PurePath
import sys
import warnings

import numpy as np

from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS

import asdf
from asdf.tags.core import NDArrayType
from asdf import AsdfFile
from asdf import schema as asdf_schema

from . import filetype
from . import fits_support
from . import properties
from . import schema as mschema
from . import validate
from .util import get_envar_as_boolean, remove_none_from_tree

from .history import HistoryList


# This minimal schema creates metadata fields that
# are accessed to be available by the core DataModel code.
_DEFAULT_SCHEMA = {
    "properties": {
        "meta": {
            "type": "object",
            "properties": {
                "filename": {
                    "type": "string",
                },
                "model_type": {
                    "type": "string",
                },
            },
        },
    },
}


class DataModel(properties.ObjectNode):
    """
    Base class of all of the data models.
    """

    schema_url = None
    """
    The schema URI to validate the model against.  If
    None, only basic validation of required metadata
    properties (filename, model_type) will occur.
    """

    def __init__(self, init=None, schema=None, memmap=False,
                 pass_invalid_values=None, strict_validation=None,
                 validate_on_assignment=None, cast_fits_arrays=True,
                 validate_arrays=False, ignore_missing_extensions=True, **kwargs):
        """
        Parameters
        ----------
        init : str, tuple, `~astropy.io.fits.HDUList`, ndarray, dict, None

            - None : Create a default data model with no shape.

            - tuple : Shape of the data array.
              Initialize with empty data array with shape specified by the.

            - file path: Initialize from the given file (FITS or ASDF)

            - readable file object: Initialize from the given file
              object

            - `~astropy.io.fits.HDUList` : Initialize from the given
              `~astropy.io.fits.HDUList`.

            - A numpy array: Used to initialize the data array

            - dict: The object model tree for the data model

        schema : dict, str (optional)
            Tree of objects representing a JSON schema, or string naming a schema.
            The schema to use to understand the elements on the model.
            If not provided, the schema associated with this class
            will be used.

        memmap : bool
            Turn memmap of FITS file on or off.  (default: False).  Ignored for
            ASDF files.

        pass_invalid_values : bool or None
            If `True`, values that do not validate the schema
            will be added to the metadata. If `False`, they will be set to `None`.
            If `None`, value will be taken from the environmental PASS_INVALID_VALUES.
            Otherwise the default value is `False`.

        strict_validation : bool or None
            If `True`, schema validation errors will generate
            an exception. If `False`, they will generate a warning.
            If `None`, value will be taken from the environmental STRICT_VALIDATION.
            Otherwise, the default value is `False`.

        validate_on_assignment : bool or None
            Defaults to 'None'.
            If `None`, value will be taken from the environmental VALIDATE_ON_ASSIGNMENT,
            defaulting to 'True' if  no environment variable is set.
            If 'True', attribute assignments are validated at the time of assignment.
            Validation errors generate warnings and values will be set to `None`.
            If 'False', schema validation occurs only once at the time of write.
            Validation errors generate warnings.

        cast_fits_arrays : bool
            If `True`, arrays will be cast to the dtype described by the schema
            when read from a FITS file.
            If `False`, arrays will be read without casting.

        validate_arrays : bool
            If `True`, arrays will be validated against ndim, max_ndim, and datatype
            validators in the schemas.

        ignore_missing_extensions : bool
            When `False`, raise warnings when a file is read that
            contains metadata about extensions that are not available.
            Defaults to `True`.

        kwargs : dict
            Additional keyword arguments passed to lower level functions. These arguments
            are generally file format-specific. Arguments of note are:

            - FITS

              skip_fits_update - bool or None
                  `True` to skip updating the ASDF tree from the FITS headers, if possible.
                  If `None`, value will be taken from the environmental SKIP_FITS_UPDATE.
                  Otherwise, the default value is `True`.
        """

        # Override value of validation parameters if not explicitly set.
        if pass_invalid_values is None:
            pass_invalid_values = get_envar_as_boolean("PASS_INVALID_VALUES",
                                                            False)
        self._pass_invalid_values = pass_invalid_values
        if strict_validation is None:
            strict_validation = get_envar_as_boolean("STRICT_VALIDATION",
                                                          False)
        if validate_on_assignment is None:
            validate_on_assignment = get_envar_as_boolean("VALIDATE_ON_ASSIGNMENT",
                                                          True)
        self._strict_validation = strict_validation
        self._ignore_missing_extensions = ignore_missing_extensions
        self._validate_on_assignment = validate_on_assignment
        self._cast_fits_arrays = cast_fits_arrays
        self._validate_arrays = validate_arrays

        kwargs.update({'ignore_missing_extensions': ignore_missing_extensions})

        # Load the schema files
        if schema is None:
            if self.schema_url is None:
                schema = _DEFAULT_SCHEMA
            else:
                # Create an AsdfFile so we can use its resolver for loading schemas
                schema = asdf_schema.load_schema(self.schema_url, resolve_references=True)

        self._schema = mschema.merge_property_trees(schema)

        # Provide the object as context to other classes and functions
        self._parent = None

        # Initialize with an empty AsdfFile instance as this is needed for
        # reading in FITS files where validate._check_value() gets called, and
        # ctx needs to have an _asdf attribute.
        self._asdf = AsdfFile()

        # Determine what kind of input we have (init) and execute the
        # proper code to intiailize the model
        self._file_references = []
        is_array = False
        is_shape = False
        shape = None

        if init is None:
            asdffile = self.open_asdf(init=None, **kwargs)

        elif isinstance(init, dict):
            asdffile = self.open_asdf(init=init, **kwargs)

        elif isinstance(init, np.ndarray):
            asdffile = self.open_asdf(init=None, **kwargs)

            shape = init.shape
            is_array = True

        elif isinstance(init, tuple):
            for item in init:
                if not isinstance(item, int):
                    raise ValueError("shape must be a tuple of ints")

            shape = init
            is_shape = True
            asdffile = self.open_asdf(init=None, **kwargs)

        elif isinstance(init, DataModel):
            asdffile = None
            self.clone(self, init)
            if not isinstance(init, self.__class__):
                self.validate()
            return

        elif isinstance(init, AsdfFile):
            asdffile = init

        elif isinstance(init, fits.HDUList):
            asdffile = fits_support.from_fits(init, self._schema, self._ctx,
                                              **kwargs)

        elif isinstance(init, (str, bytes, PurePath)):
            if isinstance(init, PurePath):
                init = str(init)
            if isinstance(init, bytes):
                init = init.decode(sys.getfilesystemencoding())
            file_type = filetype.check(init)

            if file_type == "fits":
                hdulist = fits.open(init, memmap=memmap)
                self._file_references.append(_FileReference(hdulist))
                asdffile = fits_support.from_fits(
                    hdulist, self._schema, self._ctx, **kwargs
                )

            elif file_type == "asdf":
                asdffile = self.open_asdf(init=init, **kwargs)

            else:
                # TODO handle json files as well
                raise IOError(
                        "File does not appear to be a FITS or ASDF file.")

        else:
            raise ValueError(
                "Can't initialize datamodel using {0}".format(str(type(init))))

        self._file_references.append(_FileReference(asdffile))

        # Initialize object fields as determined from the code above
        self._shape = shape
        self._instance = asdffile.tree
        self._asdf = asdffile

        # Initalize class dependent hidden fields
        self._no_asdf_extension = False

        # Instantiate the primary array of the image
        if is_array:
            primary_array_name = self.get_primary_array_name()
            if not primary_array_name:
                raise TypeError(
                    "Array passed to DataModel.__init__, but model has "
                    "no primary array in its schema")
            setattr(self, primary_array_name, init)

        # If a shape has been given, initialize the primary array.
        if is_shape:
            primary_array_name = self.get_primary_array_name()
            if not primary_array_name:
                raise TypeError(
                    "Shape passed to DataModel.__init__, but model has "
                    "no primary array in its schema")

            # Initialization occurs when the primary array is first
            # referenced. Do so now.
            getattr(self, primary_array_name)

        # initialize arrays from keyword arguments when they are present

        for attr, value in kwargs.items():
            if value is not None:
                subschema = properties._get_schema_for_property(self._schema,
                                                                attr)
                if 'datatype' in subschema:
                    setattr(self, attr, value)

        # Call hook that sets model properties
        self.on_init(init)

    @property
    def _ctx(self):
        # self._ctx is a property to avoid reference cycle that would be
        # created if we set self._ctx = self
        # a reference cycle would make DataModel difficult to garbage
        # collect and could reopen the memory leak issues fixed in
        # https://github.com/spacetelescope/stdatamodels/pull/109
        return self

    @property
    def crds_observatory(self):
        """
        Get the CRDS observatory code for this model.

        Returns
        -------
        str
        """
        raise NotImplementedError("The base DataModel class cannot be used to select best references")

    def get_crds_parameters(self):
        """
        Get the parameters used by CRDS to select references for this model.

        Returns
        -------
        dict
        """
        raise NotImplementedError("The base DataModel class cannot be used to select best references")

    @property
    def _model_type(self):
        return self.__class__.__name__

    def __repr__(self):
        buf = ['<']
        buf.append(self._model_type)

        if self.shape:
            buf.append(str(self.shape))

        try:
            filename = self.meta.filename
        except AttributeError:
            filename = None
        if filename:
            buf.append(" from ")
            buf.append(filename)
        buf.append('>')

        return "".join(buf)

    def __del__(self):
        """Ensure closure of resources when deleted."""
        self.close()

    @property
    def override_handle(self):
        """override_handle identifies in-memory models where a filepath
        would normally be used.
        """
        # Arbitrary choice to look something like crds://
        return "override://" + self.__class__.__name__

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        # This method is called by __del__, which may be invoked
        # even when the model failed to initialize.  Consequently,
        # we can't assume that any attributes have been set.
        if hasattr(self, "_file_references"):
            for file_reference in self._file_references:
                file_reference.decrement()
            # Discard the list in case close is called a second time.
            self._file_references = []

    @staticmethod
    def clone(target, source, deepcopy=False, memo=None):
        if deepcopy:
            instance = copy.deepcopy(source._instance, memo=memo)
            target._asdf = AsdfFile(instance)
            target._instance = instance
        else:
            target._asdf = source._asdf
            target._instance = source._instance
            for file_reference in source._file_references:
                file_reference.increment()
                target._file_references.append(file_reference)

        target._shape = source._shape
        target._no_asdf_extension = source._no_asdf_extension

    def copy(self, memo=None):
        """
        Returns a deep copy of this model.
        """
        result = self.__class__(init=None,
                                pass_invalid_values=self._pass_invalid_values,
                                strict_validation=self._strict_validation)
        self.clone(result, self, deepcopy=True, memo=memo)
        return result

    __copy__ = __deepcopy__ = copy

    def validate(self):
        """
        Re-validate the model instance against its schema
        """
        validate.value_change(str(self), self._instance, self._schema, self)

    def info(self, *args, **kwargs):
        return self._asdf.info(**kwargs)

    def search(self, *args, **kwargs):
        return self._asdf.search(*args, **kwargs)

    try:
        info.__doc__ = AsdfFile.info.__doc__
        search.__doc__ = AsdfFile.search.__doc__
    except AttributeError:
        pass

    def get_primary_array_name(self):
        """
        Returns the name "primary" array for this model, which
        controls the size of other arrays that are implicitly created.
        This is intended to be overridden in the subclasses if the
        primary array's name is not "data".
        """
        if properties._find_property(self._schema, 'data'):
            primary_array_name = 'data'
        else:
            primary_array_name = ''
        return primary_array_name

    def on_init(self, init):
        """
        Hook used to customize model attributes at the end of ``__init__``.

        Parameters
        ----------
        init : object
            First argument to ``__init__``.
        """
        # if the input is from a file, set the filename attribute
        if isinstance(init, str):
            self.meta.filename = os.path.basename(init)
        elif isinstance(init, fits.HDUList):
            info = init.fileinfo(0)
            if info is not None:
                filename = info.get('filename')
                if filename is not None:
                    self.meta.filename = os.path.basename(filename)

        # store the data model type, if not already set
        klass = self.__class__.__name__
        if klass != 'DataModel':
            if not self.meta.hasattr('model_type'):
                self.meta.model_type = klass

    def on_save(self, path=None):
        """
        This is a hook that is called just before saving the file.
        It can be used, for example, to update values in the metadata
        that are based on the content of the data.

        Override it in the subclass to make it do something, but don't
        forget to "chain up" to the base class, since it does things
        there, too.

        Parameters
        ----------
        path : str
            The path to the file that we're about to save to.
        """
        if isinstance(path, str):
            self.meta.filename = os.path.basename(path)

        # Enforce model_type to be the actual type of model being saved.
        self.meta.model_type = self._model_type

        # DataModel considers None to be equivalent to missing node, but
        # asdf 2.8+ writes None as null values, so we need to remove
        # Nones before proceeding with the save.
        self._instance = remove_none_from_tree(self._instance)

    def save(self, path, dir_path=None, *args, **kwargs):
        """
        Save to either a FITS or ASDF file, depending on the path.

        Parameters
        ----------
        path : string or func
            File path to save to.
            If function, it takes one argument with is
            model.meta.filename and returns the full path string.

        dir_path: string
            Directory to save to. If not None, this will override
            any directory information in the `path`

        Returns
        -------
        output_path: str
            The file path the model was saved in.
        """
        if callable(path):
            path_head, path_tail = os.path.split(path(self.meta.filename))
        else:
            path_head, path_tail = os.path.split(path)
        base, ext = os.path.splitext(path_tail)
        if isinstance(ext, bytes):
            ext = ext.decode(sys.getfilesystemencoding())

        if dir_path:
            path_head = dir_path
        output_path = os.path.join(path_head, path_tail)

        # TODO: Support gzip-compressed fits
        if ext == '.fits':
            # TODO: remove 'clobber' check once depreciated fully in astropy
            if 'clobber' not in kwargs:
                kwargs.setdefault('overwrite', True)
            self.to_fits(output_path, *args, **kwargs)
        elif ext == '.asdf':
            self.to_asdf(output_path, *args, **kwargs)
        else:
            raise ValueError("unknown filetype {0}".format(ext))

        return output_path

    @staticmethod
    def open_asdf(init=None,
                  ignore_version_mismatch=True,
                  ignore_unrecognized_tag=False,
                  **kwargs):
        """
        Open an asdf object from a filename or create a new asdf object
        """
        if isinstance(init, str):
            asdffile = asdf.open(init,
                                 ignore_version_mismatch=ignore_version_mismatch,
                                 ignore_unrecognized_tag=ignore_unrecognized_tag)

        else:
            asdffile = AsdfFile(init,
                            ignore_version_mismatch=ignore_version_mismatch,
                            ignore_unrecognized_tag=ignore_unrecognized_tag
                            )
        return asdffile

    @classmethod
    def from_asdf(cls, init, schema=None, **kwargs):
        """
        Load a data model from an ASDF file.

        Parameters
        ----------
        init : str, file object, `~asdf.AsdfFile`
            - str : file path: initialize from the given file
            - readable file object: Initialize from the given file object
            - `~asdf.AsdfFile` : Initialize from the given`~asdf.AsdfFile`.
        schema :
            Same as for `__init__`
        kwargs : dict
            Aadditional arguments passed to lower level functions

        Returns
        -------
        model : `~jwst.datamodels.DataModel` instance
            A data model.
        """
        return cls(init, schema=schema, **kwargs)


    def to_asdf(self, init, *args, **kwargs):
        """
        Write a data model to an ASDF file.

        Parameters
        ----------
        init : file path or file object
        args : tuple, list
            Additional positional arguments passed to `~asdf.AsdfFile.write_to`.
        kwargs : dict
            Any additional keyword arguments are passed along to
            `~asdf.AsdfFile.write_to`.
        """
        self.on_save(init)
        self.validate()
        asdffile = self.open_asdf(self._instance, **kwargs)
        asdffile.write_to(init, *args, **kwargs)

    @classmethod
    def from_fits(cls, init, schema=None, **kwargs):
        """
        Load a model from a FITS file.

        Parameters
        ----------
        init : file path, file object, astropy.io.fits.HDUList
            - file path: Initialize from the given file
            - readable file object: Initialize from the given file object
            - astropy.io.fits.HDUList: Initialize from the given
              `~astropy.io.fits.HDUList`.

        schema : dict, str
            Same as for `__init__`

        kwargs : dict
            Aadditional arguments passed to lower level functions.

        Returns
        -------
        model : `~jwst.datamodels.DataModel`
            A data model.
        """
        return cls(init, schema=schema, **kwargs)

    def to_fits(self, init, *args, **kwargs):
        """
        Write a data model to a FITS file.

        Parameters
        ----------
        init : file path or file object

        args, kwargs
            Any additional arguments are passed along to
            `astropy.io.fits.writeto`.
        """
        self.on_save(init)

        hdulist = fits_support.to_fits(self._instance, self._schema)
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', message='Card is too long')
            if self._no_asdf_extension:
                # For some old files that were written out before the
                # _no_asdf_extension existed, these will have an ASDF
                # extension, which may get passed along through extra_fits.
                # Avoid this.
                if "ASDF" in hdulist:
                    del hdulist["ASDF"]
            hdulist.writeto(init, *args, **kwargs)

    @property
    def shape(self):
        if self._shape is None:
            primary_array_name = self.get_primary_array_name()
            if primary_array_name and self.hasattr(primary_array_name):
                primary_array = getattr(self, primary_array_name)
                self._shape = primary_array.shape
        return self._shape

    def __setattr__(self, attr, value):
        if attr in frozenset(("shape", "history", "_extra_fits", "schema")):
            object.__setattr__(self, attr, value)
        else:
            properties.ObjectNode.__setattr__(self, attr, value)

    def extend_schema(self, new_schema):
        """
        Extend the model's schema using the given schema, by combining
        it in an "allOf" array.

        Parameters
        ----------
        new_schema : dict
            Schema tree.
        """
        schema = {'allOf': [self._schema, new_schema]}
        self._schema = mschema.merge_property_trees(schema)
        self.validate()
        return self

    def add_schema_entry(self, position, new_schema):
        """
        Extend the model's schema by placing the given new_schema at
        the given dot-separated position in the tree.

        Parameters
        ----------
        position : str
            Dot separated string indicating the position, e.g. ``meta.instrument.name``.
        new_schema : dict
            Schema tree.
        """
        parts = position.split('.')
        schema = new_schema
        for part in parts[::-1]:
            schema = {'type': 'object', 'properties': {part: schema}}
        return self.extend_schema(schema)

    # return_result retained for backward compatibility
    def find_fits_keyword(self, keyword, return_result=True):
        """
        Utility function to find a reference to a FITS keyword in this
        model's schema.  This is intended for interactive use, and not
        for use within library code.

        Parameters
        ----------
        keyword : str
            A FITS keyword name.

        Returns
        -------
        locations : list of str
            If `return_result` is `True`, a list of the locations in
            the schema where this FITS keyword is used.  Each element
            is a dot-separated path.
        """
        from . import schema
        return schema.find_fits_keyword(self.schema, keyword)

    def search_schema(self, substring):
        """
        Utility function to search the metadata schema for a
        particular phrase.

        This is intended for interactive use, and not for use within
        library code.

        The searching is case insensitive.

        Parameters
        ----------
        substring : str
            The substring to search for.

        Returns
        -------
        locations : list of tuples
        """
        from . import schema
        return schema.search_schema(self.schema, substring)

    def __getitem__(self, key):
        """
        Get a metadata value using a dotted name.
        """
        assert isinstance(key, str)
        meta = self
        for part in key.split('.'):
            try:
                meta = getattr(meta, part)
            except AttributeError:
                raise KeyError(repr(key))
        return meta

    def __setitem__(self, key, value):
        """
        Set a metadata value using a dotted name.
        """
        assert isinstance(key, str)
        meta = self
        parts = key.split('.')
        for part in parts[:-1]:
            try:
                part = int(part)
            except ValueError:
                try:
                    meta = getattr(meta, part)
                except AttributeError:
                    raise KeyError(repr(key))
            else:
                meta = meta[part]

        part = parts[-1]
        try:
            part = int(part)
        except ValueError:
            setattr(meta, part, value)
        else:
            meta[part] = value

    def items(self):
        """
        Iterates over all of the schema items in a flat way.

        Each element is a pair (`key`, `value`).  Each `key` is a
        dot-separated name.  For example, the schema element
        `meta.observation.date` will end up in the result as::

            ("meta.observation.date": "2012-04-22T03:22:05.432")
        """
        def recurse(tree, path=[]):
            if isinstance(tree, dict):
                for key, val in tree.items():
                    for x in recurse(val, path + [key]):
                        yield x
            elif isinstance(tree, (list, tuple)):
                for i, val in enumerate(tree):
                    for x in recurse(val, path + [i]):
                        yield x
            elif tree is not None:
                yield ('.'.join(str(x) for x in path), tree)

        for x in recurse(self._instance):
            yield x

    def keys(self):
        """
        Iterates over all of the schema keys in a flat way.

        Each result of the iterator is a `key`.  Each `key` is a
        dot-separated name.  For example, the schema element
        `meta.observation.date` will end up in the result as the
        string `"meta.observation.date"`.
        """
        for key, val in self.items():
            yield key

    def values(self):
        """
        Iterates over all of the schema values in a flat way.
        """
        for key, val in self.items():
            yield val

    def update(self, d, only=None, extra_fits=False):
        """
        Updates this model with the metadata elements from another model.

        Note: The ``update`` method skips a WCS object, if present.

        Parameters
        ----------
        d : `~jwst.datamodels.DataModel` or dictionary-like object
            The model to copy the metadata elements from. Can also be a
            dictionary or dictionary of dictionaries or lists.
        only: str, None
            Update only the named hdu, e.g. ``only='PRIMARY'``. Can either be
            a string or list of hdu names. Default is to update all the hdus.
        extra_fits : boolean
            Update from ``extra_fits``.  Default is False.
        """
        def hdu_keywords_from_data(d, path, hdu_keywords):
            # Walk tree and add paths to keywords to hdu keywords
            if isinstance(d, dict):
                for key, val in d.items():
                    if len(path) > 0 or key != 'extra_fits':
                        hdu_keywords_from_data(val, path + [key], hdu_keywords)
            elif isinstance(d, list):
                for key, val in enumerate(d):
                    hdu_keywords_from_data(val, path + [key], hdu_keywords)
            elif isinstance(d, np.ndarray):
                # skip data arrays
                pass
            else:
                hdu_keywords.append(path)

        def hdu_keywords_from_schema(subschema, path, combiner, ctx, recurse):
            # Add path to keyword to hdu_keywords if in list of hdu names
            if 'fits_keyword' in subschema:
                fits_hdu = subschema.get('fits_hdu', 'PRIMARY')
                if fits_hdu in hdu_names:
                    ctx.append(path)

        def hdu_names_from_schema(subschema, path, combiner, ctx, recurse):
            # Build a set of hdu names from the schema
            hdu_name = subschema.get('fits_hdu')
            if hdu_name:
                hdu_names.add(hdu_name)

        def included(cursor, part):
            # Test if part is in the cursor
            if isinstance(part, int):
                return part >= 0 and part < len(cursor)
            else:
                return part in cursor

        def set_hdu_keyword(this_cursor, that_cursor, path):
            # Copy an element pointed to by path from that to this
            part = path.pop(0)
            if not included(that_cursor, part):
                return
            if len(path) == 0:
                this_cursor[part] = copy.deepcopy(that_cursor[part])
            else:
                that_cursor = that_cursor[part]
                if not included(this_cursor, part):
                    if isinstance(path[0], int):
                        if isinstance(part, int):
                            this_cursor.append([])
                        else:
                            this_cursor[part] = []
                    else:
                        if isinstance(part, int):
                            this_cursor.append({})
                        elif isinstance(that_cursor, list):
                            this_cursor[part] = []
                        else:
                            this_cursor[part] = {}
                this_cursor = this_cursor[part]
                set_hdu_keyword(this_cursor, that_cursor, path)

        def protected_keyword(path):
            # Some keywords are protected and
            # should not be copied frpm the other image
            if len(path) == 2:
                if path[0] == 'meta':
                    if path[1] in ('date', 'model_type'):
                        return True
            return False
        # Get the list of hdu names from the model so that updates
        # are limited to those hdus

        if only is not None:
            if isinstance(only, str):
                hdu_names = set([only])
            else:
                hdu_names = set(list(only))
        else:
            hdu_names = set(['PRIMARY'])
            mschema.walk_schema(self._schema, hdu_names_from_schema, hdu_names)

        # Get the paths to all the keywords that will be updated from

        hdu_keywords = []
        if isinstance(d, DataModel):
            schema = d._schema
            d = d._instance
            mschema.walk_schema(schema, hdu_keywords_from_schema, hdu_keywords)
        else:
            path = []
            hdu_keywords_from_data(d, path, hdu_keywords)

        # Perform the updates to the keywords mentioned in the schema
        for path in hdu_keywords:
            if not protected_keyword(path):
                set_hdu_keyword(self._instance, d, path)

        # Update from extra_fits as well, if indicated
        if extra_fits:
            for hdu_name in hdu_names:
                path = ['extra_fits', hdu_name, 'header']
                set_hdu_keyword(self._instance, d, path)

        self.validate()

    def to_flat_dict(self, include_arrays=True):
        """
        Returns a dictionary of all of the schema items as a flat dictionary.

        Each dictionary key is a dot-separated name.  For example, the
        schema element `meta.observation.date` will end up in the
        dictionary as::

            { "meta.observation.date": "2012-04-22T03:22:05.432" }

        """
        def convert_val(val):
            if isinstance(val, datetime.datetime):
                return val.isoformat()
            elif isinstance(val, Time):
                return str(val)
            return val

        if include_arrays:
            return dict((key, convert_val(val)) for (key, val) in self.items())
        else:
            return dict((key, convert_val(val)) for (key, val) in self.items()
                        if not isinstance(val, (np.ndarray, NDArrayType)))

    @property
    def schema(self):
        return self._schema

    def get_fileext(self):
        return 'fits'

    @property
    def history(self):
        """
        Get the history as a list of entries
        """
        return HistoryList(self._asdf)

    @history.setter
    def history(self, values):
        """
        Set a history entry.

        Parameters
        ----------
        values : list
            For FITS files this should be a list of strings.
            For ASDF files use a list of ``HistoryEntry`` object. It can be created
            with `~jwst.datamodels.util.create_history_entry`.

        """
        entries = self.history
        entries.clear()
        entries.extend(values)

    def get_fits_wcs(self, hdu_name='SCI', hdu_ver=1, key=' '):
        """
        Get a `astropy.wcs.WCS` object created from the FITS WCS
        information in the model.

        Note that modifying the returned WCS object will not modify
        the data in this model.  To update the model, use `set_fits_wcs`.

        Parameters
        ----------
        hdu_name : str, optional
            The name of the HDU to get the WCS from.  This must use
            named HDU's, not numerical order HDUs. To get the primary
            HDU, pass ``'PRIMARY'``.

        key : str, optional
            The name of a particular WCS transform to use.  This may
            be either ``' '`` or ``'A'``-``'Z'`` and corresponds to
            the ``"a"`` part of the ``CTYPEia`` cards.  *key* may only
            be provided if *header* is also provided.

        hdu_ver: int, optional
            The extension version. Used when there is more than one
            extension with the same name. The default value, 1,
            is the first.

        Returns
        -------
        wcs : `astropy.wcs.WCS` or `pywcs.WCS` object
            The type will depend on what libraries are installed on
            this system.
        """
        hdulist = fits_support.to_fits(self._instance, self._schema)
        hdu = fits_support.get_hdu(hdulist, hdu_name, index=hdu_ver-1)
        header = hdu.header
        return WCS(header, key=key, relax=True, fix=True)

    def set_fits_wcs(self, wcs, hdu_name='SCI'):
        """
        Sets the FITS WCS information on the model using the given
        `astropy.wcs.WCS` object.

        Note that the "key" of the WCS is stored in the WCS object
        itself, so it can not be set as a parameter to this method.

        Parameters
        ----------
        wcs : `astropy.wcs.WCS` or `pywcs.WCS` object
            The object containing FITS WCS information

        hdu_name : str, optional
            The name of the HDU to set the WCS from.  This must use
            named HDU's, not numerical order HDUs.  To set the primary
            HDU, pass ``'PRIMARY'``.
        """
        header = wcs.to_header()
        if hdu_name == 'PRIMARY':
            hdu = fits.PrimaryHDU(header=header)
        else:
            hdu = fits.ImageHDU(name=hdu_name, header=header)
        hdulist = fits.HDUList([hdu])

        ff = fits_support.from_fits(hdulist, self._schema, self._ctx,
                                    ignore_missing_extensions=self._ignore_missing_extensions)

        self._instance = properties.merge_tree(self._instance, ff.tree)

    # --------------------------------------------------------
    # These two method aliases are here for astropy.registry
    # compatibility and should not be called directly
    # --------------------------------------------------------

    read = __init__

    def write(self, path, *args, **kwargs):
        self.save(path, *args, **kwargs)

    def getarray_noinit(self, attribute):
        """Retrieve array but without initilization

        Arrays initialize when directly referenced if they had
        not previously been initialized. This circumvents the
        initialization and instead raises `AttributeError`.

        Parameters
        ----------
        attribute : str
            The attribute to retrieve.

        Returns
        -------
        value : object
           The value of the attribute.

        Raises
        ------
        AttributeError
            If the attribute does not exist.
        """
        if attribute in self.instance:
            return getattr(self, attribute)
        raise AttributeError(f'{self} has no attribute "{attribute}"')


class _FileReference:
    """
    Reference counter for open file pointers managed by
    DataModel.  Once decremented to zero the file will
    be closed.
    """

    def __init__(self, file):
        self._file = file
        self._count = 1

    def increment(self):
        self._count += 1

    def decrement(self):
        if self._count <= 0:
            return

        self._count -= 1
        if self._count <= 0:
            self._file.close()
            self._file = None



from .model_base import JwstDataModel, DataModel
from .abvega_offset import ABVegaOffsetModel
from .amilg import AmiLgModel
from .amioi import AmiOIModel
from .apcorr import FgsImgApcorrModel, MirImgApcorrModel
from .apcorr import NrcImgApcorrModel, NisImgApcorrModel
from .apcorr import MirLrsApcorrModel, MirMrsApcorrModel
from .apcorr import NrcWfssApcorrModel, NisWfssApcorrModel
from .apcorr import NrsMosApcorrModel, NrsIfuApcorrModel, NrsFsApcorrModel
from .asn import AsnModel
from .barshadow import BarshadowModel
from .combinedspec import CombinedSpecModel
from .contrast import ContrastModel
from .cube import CubeModel
from .dark import DarkModel
from .darkMIRI import DarkMIRIModel
from .drizpars import DrizParsModel
from .drizproduct import DrizProductModel
from .extract1dimage import Extract1dImageModel
from .extract1d_spec import Extract1dIFUModel
from .flat import FlatModel
from .fringe import FringeModel
from .fringefreq import FringeFreqModel
from .gain import GainModel
from .gls_rampfit import GLS_RampFitModel
from .guider import GuiderRawModel, GuiderCalModel
from .ifucube import IFUCubeModel
from .ifucubepars import NirspecIFUCubeParsModel, MiriIFUCubeParsModel
from .ifuimage import IFUImageModel
from .mrsptcorr import MirMrsPtCorrModel
from .image import ImageModel
from .ipc import IPCModel
from .irs2 import IRS2Model
from .lastframe import LastFrameModel
from .level1b import Level1bModel
from .linearity import LinearityModel
from .mask import MaskModel
from .ramp import MIRIRampModel
from .mrsxartcorr import MirMrsXArtCorrModel
from .multicombinedspec import MultiCombinedSpecModel
from .multiexposure import MultiExposureModel
from .multiextract1d import MultiExtract1dImageModel
from .multiprod import MultiProductModel
from .multislit import MultiSlitModel
from .multispec import MultiSpecModel
from .nirspec_flat import NirspecFlatModel, NirspecQuadFlatModel
from .outlierpars import OutlierParsModel
from .outlierifuoutput import OutlierIFUOutputModel
from .pathloss import PathlossModel, MirLrsPathlossModel
from .persat import PersistenceSatModel
from .photom import FgsImgPhotomModel
from .photom import MirImgPhotomModel, MirLrsPhotomModel, MirMrsPhotomModel
from .photom import NrcImgPhotomModel, NrcWfssPhotomModel
from .photom import NisImgPhotomModel, NisSossPhotomModel, NisWfssPhotomModel
from .photom import NrsFsPhotomModel, NrsMosPhotomModel
from .pixelarea import PixelAreaModel, NirspecSlitAreaModel, NirspecMosAreaModel, NirspecIfuAreaModel
from .psfmask import PsfMaskModel
from .quad import QuadModel
from .ramp import RampModel
from .rampfitoutput import RampFitOutputModel
from .readnoise import ReadnoiseModel
from .reference import ReferenceFileModel, ReferenceImageModel, ReferenceCubeModel, ReferenceQuadModel
from .reset import ResetModel
from .resolution import ResolutionModel, MiriResolutionModel
from .rscd import RSCDModel
from .saturation import SaturationModel
from .segmap import SegmentationMapModel
from .slit import SlitModel, SlitDataModel
from .sossextractmodel import SossExtractModel
from .sosswavegrid import SossWaveGridModel
from .spec import SpecModel
from .speckernel import SpecKernelModel
from .specprofile import SpecProfileModel, SpecProfileSingleModel
from .spectrace import SpecTraceModel, SpecTraceSingleModel
from .straylight import StrayLightModel
from .superbias import SuperBiasModel
from .throughput import ThroughputModel
from .trapdensity import TrapDensityModel
from .trappars import TrapParsModel
from .trapsfilled import TrapsFilledModel
from .tsophot import TsoPhotModel
from .wavemap import WaveMapModel, WaveMapSingleModel
from .wcs_ref_models import (DistortionModel, DistortionMRSModel, SpecwcsModel,
                             RegionsModel, WavelengthrangeModel, CameraModel, CollimatorModel, OTEModel,
                             FOREModel, FPAModel, IFUPostModel, IFUFOREModel, IFUSlicerModel, MSAModel,
                             FilteroffsetModel, DisperserModel, NIRCAMGrismModel, NIRISSGrismModel,
                             WaveCorrModel)
from .wfssbkg import WfssBkgModel
from .util import open


__all__ = [
    'open',
    'DataModel', 'JwstDataModel',
    'ABVegaOffsetModel',
    'AmiLgModel',
    'AmiOIModel',
    'FgsImgApcorrModel', 'MirImgApcorrModel', 'NrcImgApcorrModel', 'NisImgApcorrModel',
    'MirLrsApcorrModel', 'MirMrsApcorrModel', 'NrcWfssApcorrModel', 'NisWfssApcorrModel',
    'NrsMosApcorrModel', 'NrsFsApcorrModel', 'NrsIfuApcorrModel',
    'AsnModel',
    'BarshadowModel', 'CameraModel', 'CollimatorModel',
    'CombinedSpecModel', 'ContrastModel', 'CubeModel',
    'DarkModel', 'DarkMIRIModel',
    'DisperserModel', 'DistortionModel', 'DistortionMRSModel',
    'DrizParsModel',
    'DrizProductModel',
    'Extract1dImageModel',
    'Extract1dIFUModel',
    'FilteroffsetModel',
    'FlatModel', 'NirspecFlatModel', 'NirspecQuadFlatModel',
    'FOREModel', 'FPAModel',
    'FringeModel', 'FringeFreqModel', 'GainModel', 'GLS_RampFitModel',
    'GuiderRawModel', 'GuiderCalModel',
    'IFUCubeModel',
    'NirspecIFUCubeParsModel', 'MiriIFUCubeParsModel', 'MirMrsPtCorrModel',
    'MirMrsXArtCorrModel',
    'IFUFOREModel', 'IFUImageModel', 'IFUPostModel', 'IFUSlicerModel',
    'ImageModel', 'IPCModel', 'IRS2Model', 'LastFrameModel', 'Level1bModel',
    'LinearityModel', 'MaskModel', 'MSAModel',
    'MultiCombinedSpecModel', 'MultiExposureModel',
    'MultiExtract1dImageModel', 'MultiSlitModel',
    'MultiProductModel',
    'MultiSpecModel',
    'NIRCAMGrismModel', 'NIRISSGrismModel',
    'OTEModel',
    'OutlierParsModel','OutlierIFUOutputModel',
    'PathlossModel', 'MirLrsPathlossModel',
    'PersistenceSatModel',
    'PixelAreaModel', 'NirspecSlitAreaModel', 'NirspecMosAreaModel', 'NirspecIfuAreaModel',
    'FgsImgPhotomModel',
    'MirImgPhotomModel', 'MirLrsPhotomModel', 'MirMrsPhotomModel',
    'NrcImgPhotomModel', 'NrcWfssPhotomModel',
    'NisImgPhotomModel', 'NisSossPhotomModel', 'NisWfssPhotomModel',
    'NrsFsPhotomModel', 'NrsMosPhotomModel',
    'PsfMaskModel',
    'QuadModel', 'RampModel', 'MIRIRampModel',
    'RampFitOutputModel', 'ReadnoiseModel',
    'ReferenceFileModel', 'ReferenceCubeModel', 'ReferenceImageModel', 'ReferenceQuadModel',
    'RegionsModel', 'ResetModel',
    'ResolutionModel', 'MiriResolutionModel',
    'RSCDModel', 'SaturationModel', 'SlitDataModel', 'SlitModel', 'SpecModel',
    'SegmentationMapModel',
    'SossExtractModel',
    'SossWaveGridModel',
    'SpecKernelModel',
    'SpecProfileModel', 'SpecProfileSingleModel',
    'SpecTraceModel', 'SpecTraceSingleModel',
    'SpecwcsModel',
    'StrayLightModel', 'SuperBiasModel',
    'ThroughputModel',
    'TrapDensityModel', 'TrapParsModel', 'TrapsFilledModel',
    'TsoPhotModel',
    'WavelengthrangeModel', 'WaveCorrModel',
    'WaveMapModel', 'WaveMapSingleModel',
    'WfssBkgModel']


_all_models = __all__[1:]
_local_dict = locals()
_defined_models = {k: _local_dict[k] for k in _all_models}

# Modules that are not part of public API
_private_modules = ["conftest", "integration"]



from .reference import ReferenceFileModel

__all__ = ['FgsImgApcorrModel', 'MirImgApcorrModel',
           'NrcImgApcorrModel', 'NisImgApcorrModel',
           'MirLrsApcorrModel', 'MirMrsApcorrModel',
           'NrcWfssApcorrModel', 'NisWfssApcorrModel',
           'NrsMosApcorrModel', 'NrsIfuApcorrModel', 'NrsFsApcorrModel']


class FgsImgApcorrModel(ReferenceFileModel):
    """
    A data model for FGS imaging apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - eefraction: float32
        - radius: float32
        - apcorr: float32
        - skyin: float32
        - skyout: float32

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/fgsimg_apcorr.schema"


class MirImgApcorrModel(ReferenceFileModel):
    """
    A data model for MIRI imaging apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - subarray: str[15]
        - eefraction: float32
        - radius: float32
        - apcorr: float32
        - skyin: float32
        - skyout: float32

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/mirimg_apcorr.schema"


class MirLrsApcorrModel(ReferenceFileModel):
    """
    A data model for MIRI LRS apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - subarray: str[15]
        - wavelength: float32 1D array
        - nelem_wl: int16
        - size: uint8 1D array
        - nelem_size: int16
        - apcorr: float32 2D array
        - apcorr_err: float32 2D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/mirlrs_apcorr.schema"


class MirMrsApcorrModel(ReferenceFileModel):
    """
    A data model for MIRI MRS apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - wavelength: float32 1D array
        - radius: float32 2D array
        - apcorr: float32 2D array
        - apcorr_err: float32 2D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/mirmrs_apcorr.schema"


class NrcImgApcorrModel(ReferenceFileModel):
    """
    A data model for NIRCam imaging apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - pupil: str[15]
        - eefraction: float32
        - radius: float32
        - apcorr: float32
        - skyin: float32
        - skyout: float32

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrcimg_apcorr.schema"


class NrcWfssApcorrModel(ReferenceFileModel):
    """
    A data model for NIRCam WFSS apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - pupil: str[15]
        - wavelength: float32 1D array
        - nelem_wl: int16
        - size: uint8 1D array
        - nelem_size: int16
        - apcorr: float32 2D array
        - apcorr_err: float32 2D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrcwfss_apcorr.schema"


class NisImgApcorrModel(ReferenceFileModel):
    """
    A data model for NIRISS imaging apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - pupil: str[15]
        - eefraction: float32
        - radius: float32
        - apcorr: float32
        - skyin: float32
        - skyout: float32

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nisimg_apcorr.schema"


class NisWfssApcorrModel(ReferenceFileModel):
    """
    A data model for NIRISS WFSS apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - pupil: str[15]
        - wavelength: float32 1D array
        - nelem_wl: int16
        - size: uint8 1D array
        - nelem_size: int16
        - apcorr: float32 2D array
        - apcorr_err: float32 2D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/niswfss_apcorr.schema"


class NrsMosApcorrModel(ReferenceFileModel):
    """
    A data model for NIRSpec MOS apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - grating: str[15]
        - wavelength: float32 1D array
        - nelem_wl: int16
        - size: float32 2D array
        - nelem_size: int16
        - pixphase: float32 1D array
        - apcorr: float32 3D array
        - apcorr_err: float32 3D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrsmos_apcorr.schema"


class NrsIfuApcorrModel(ReferenceFileModel):
    """
    A data model for NIRSpec IFU apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - grating: str[15]
        - wavelength: float32 1D array
        - radius: float32 3D array
        - apcorr: float32 3D array
        - apcorr_err: float32 3D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrsifu_apcorr.schema"


class NrsFsApcorrModel(ReferenceFileModel):
    """
    A data model for NIRSpec Fixed-Slit apcorr reference files.

    Parameters
    __________
    apcorr_table : numpy table
        Aperture correction factors table
        A table-like object containing row selection criteria made up
        of instrument mode parameters and aperture correction
        factors associated with those modes.

        - filter: str[12]
        - grating: str[15]
        - slit: str[15]
        - wavelength: float32 1D array
        - nelem_wl: int16
        - size: float32 2D array
        - nelem_size: int16
        - pixphase: float32 1D array
        - apcorr: float32 3D array
        - apcorr_err: float32 3D array

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrsfs_apcorr.schema"



from .reference import ReferenceFileModel


__all__ = ['SpecProfileModel', 'SpecProfileSingleModel']


class SpecProfileModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS spectral profile reference files.

    This model has a special member `profile` that can be used to
    deal with an entire spectral profile at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import SpecProfileSingleModel
       >>> specprofile_model = SpecProfileModel()
       >>> specprofile_model.profile.append(SpecProfileSingleModel())
       >>> specprofile_model.profile[0] # doctest: +SKIP
       <SpecProfileSingleModel>

    If `init` is a `SpecProfileSingleModel` instance, an empty `SpecProfileSingleModel`
    will be created and assigned to attribute `profile[0]`, and the `data`
    attribute from the input `SpecProfileSingleModel` instance will be copied to
    the first element of `profile`.  `SpecProfileSingleModel` objects can be appended
    to the `profile` attribute by using its `append` method.

    Parameters
    __________
    profile.items.data : numpy array
         Spectral profile data

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/specprofile.schema"

    def __init__(self, init=None, **kwargs):

        if isinstance(init, SpecProfileSingleModel):
            super(SpecProfileModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.profile.append(self.profile.item())
            self.profile[0].data = init.data

        super(SpecProfileModel, self).__init__(init=init, **kwargs)


class SpecProfileSingleModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS spectral profile data.

    Parameters
    __________
    data : numpy float32 array
         Spectral profile values
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/specprofilesingle.schema"



from .model_base import JwstDataModel
from stdatamodels.jwst.library.basic_utils import deprecate_class

__all__ = ['RampModel', 'MIRIRampModel']


class RampModel(JwstDataModel):
    """
    A data model for 4D ramps.

    Parameters
    __________
    data : numpy float32 array
         The science data

    pixeldq : numpy uint32 array
         2-D data quality array for all planes

    groupdq : numpy uint8 array
         4-D data quality array for each plane

    err : numpy float32 array
         Error array

    zeroframe : numpy float32 array
         Zeroframe array

    group : numpy table
         group parameters table

    int_times : numpy table
         table of times for each integration

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/ramp.schema"

    def __init__(self, init=None, **kwargs):
        super(RampModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.pixeldq = self.pixeldq
        self.groupdq = self.groupdq
        self.err = self.err


@deprecate_class(RampModel)
class MIRIRampModel:
    """A data model for 4D MIRI ramps.

    This model has been deprecated. Please use `RampModel` instead.
    """



from .model_base import JwstDataModel


__all__ = ['ContrastModel']


class ContrastModel(JwstDataModel):
    """
    A data model for coronagraphic contrast curve files.

    Parameters
    __________
    contrast_table : numpy table
         Contrast curve table
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/contrast.schema"



import os
import warnings

from asdf.exceptions import ValidationError
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from numpy.testing import assert_allclose, assert_array_equal
import numpy as np
import pytest

from stdatamodels.jwst.datamodels import (JwstDataModel, ImageModel, MaskModel, AsnModel,
                                          MultiSlitModel, SlitModel, DataModel,
                                          DrizProductModel, MultiProductModel, MIRIRampModel,
                                          SlitDataModel, IFUImageModel, ABVegaOffsetModel)
from stdatamodels.jwst import datamodels
from stdatamodels.jwst.datamodels import _defined_models as defined_models

ROOT_DIR = os.path.join(os.path.dirname(__file__), 'data')
FITS_FILE = os.path.join(ROOT_DIR, 'test.fits')
ASN_FILE = os.path.join(ROOT_DIR, 'association.json')


@pytest.fixture
def make_models(tmp_path):
    """Create basic models

    Returns
    -------
    path_just_fits, path_model : (str, str)
        `path_just_fits` is a FITS file of `JwstDataModel` without the ASDF extension.
        `path_model` is a FITS file of `JwstDataModel` with the ASDF extension.
    """
    path_just_fits = tmp_path / 'just_fits.fits'
    path_model = tmp_path / 'model.fits'
    primary_hdu = fits.PrimaryHDU()
    primary_hdu.header['EXP_TYPE'] = 'NRC_IMAGE'
    primary_hdu.header['DATAMODL'] = "JwstDataModel"
    hduls = fits.HDUList([primary_hdu])
    hduls.writeto(path_just_fits)
    model = JwstDataModel(hduls)
    model.save(path_model)
    return {
        'just_fits': path_just_fits,
        'model': path_model
    }


def test_init_from_pathlib(tmp_path):
    """Test initializing model from a Path object"""
    path = tmp_path / "pathlib.fits"
    model1 = datamodels.ImageModel((50, 50))
    model1.save(path)
    with datamodels.open(path) as model:
        # Test is basically, did we open the model?
        assert isinstance(model, ImageModel)


@pytest.mark.parametrize('which_file, skip_fits_update, expected_exp_type',
                         [
                             ('just_fits', None, 'FGS_DARK'),
                             ('just_fits', False, 'FGS_DARK'),
                             ('just_fits', True, 'FGS_DARK'),
                             ('model', None, 'FGS_DARK'),
                             ('model', False, 'FGS_DARK'),
                             ('model', True, 'NRC_IMAGE')
                         ]
                         )
@pytest.mark.parametrize('use_env', [False, True])
def test_skip_fits_update(jail_environ,
                          use_env,
                          make_models,
                          which_file,
                          skip_fits_update,
                          expected_exp_type):
    """Test skip_fits_update setting"""
    # Setup the FITS file, modifying a header value
    path = make_models[which_file]
    with fits.open(path) as hduls:
        hduls[0].header['exp_type'] = 'FGS_DARK'

        # Decide how to skip. If using the environmental,
        # set that and pass None to the open function.
        try:
            del os.environ['SKIP_FITS_UPDATE']
        except KeyError:
            # No need to worry, environmental doesn't exist anyways
            pass
        if use_env:
            if skip_fits_update is not None:
                os.environ['SKIP_FITS_UPDATE'] = str(skip_fits_update)
                skip_fits_update = None

        with datamodels.open(hduls, skip_fits_update=skip_fits_update) as model:
            assert model.meta.exposure.type == expected_exp_type


def test_asnmodel_table_size_zero():
    with AsnModel() as dm:
        assert len(dm.asn_table) == 0


def test_imagemodel():
    shape = (10, 10)
    with ImageModel(shape) as dm:
        assert dm.data.shape == shape
        assert dm.err.shape == shape
        assert dm.dq.shape == shape
        assert dm.data.mean() == 0.0
        assert dm.err.mean() == 0.0
        assert dm.dq.mean() == 0.0
        assert dm.zeroframe.shape == shape
        assert dm.area.shape == shape
        assert dm.pathloss_point.shape == shape
        assert dm.pathloss_uniform.shape == shape


def test_model_with_nonstandard_primary_array():
    class NonstandardPrimaryArrayModel(JwstDataModel):
        schema_url = os.path.join(ROOT_DIR, "nonstandard_primary_array.schema.yaml")

        # The wavelength array is the primary array.
        # Try commenting this function out and the problem goes away.
        def get_primary_array_name(self):
            return 'wavelength'

    m = NonstandardPrimaryArrayModel((10,))
    assert 'wavelength' in list(m.keys())
    assert m.wavelength.sum() == 0


def test_image_with_extra_keyword_to_multislit(tmp_path):
    path = tmp_path / "extra_keyword_image.fits"
    path2 = tmp_path / "extra_keyword_multislit.fits"
    with ImageModel((32, 32)) as im:
        im.save(path)

    from astropy.io import fits
    with fits.open(path, mode="update", memmap=False) as hdulist:
        hdulist[1].header['BUNIT'] = 'x'

    with ImageModel(path) as im:
        with MultiSlitModel() as ms:
            ms.update(im)
            for i in range(3):
                ms.slits.append(ImageModel((4, 4)))
            assert len(ms.slits) == 3

            ms.save(path2)

    with MultiSlitModel(path2) as ms:
        assert len(ms.slits) == 3
        for slit in ms.slits:
            assert slit.data.shape == (4, 4)


@pytest.fixture
def datamodel_for_update(tmp_path):
    """Provide ImageModel with one keyword each defined in PRIMARY and SCI
    extensions from the schema, and one each not in the schema"""
    path = tmp_path / "old.fits"
    with ImageModel((5, 5)) as im:
        # Add schema keywords, one to each extension
        im.meta.telescope = "JWST"
        im.meta.wcsinfo.crval1 = 5

        im.save(path)
    # Add non-schema keywords that will get dumped in the extra_fits attribute
    with fits.open(path, mode="update") as hdulist:
        hdulist["PRIMARY"].header["FOO"] = "BAR"
        hdulist["SCI"].header["BAZ"] = "BUZ"

    return path


@pytest.mark.parametrize("extra_fits", [True, False])
@pytest.mark.parametrize("only", [None, "PRIMARY", "SCI"])
def test_update_from_datamodel(tmp_path, datamodel_for_update, only, extra_fits):
    """Test update method does not update from extra_fits unless asked"""
    path = tmp_path / "new.fits"
    with ImageModel((5, 5)) as newim:
        with ImageModel(datamodel_for_update) as oldim:

            # Verify the fixture returns keywords we expect
            assert oldim.meta.telescope == "JWST"
            assert oldim.meta.wcsinfo.crval1 == 5
            assert oldim.extra_fits.PRIMARY.header == [['FOO', 'BAR', '']]
            assert oldim.extra_fits.SCI.header == [['BAZ', 'BUZ', '']]

            newim.update(oldim, only=only, extra_fits=extra_fits)
        newim.save(path)

    with fits.open(path) as hdulist:
        if extra_fits:
            if only == "PRIMARY":
                assert "TELESCOP" in hdulist["PRIMARY"].header
                assert "CRVAL1" not in hdulist["SCI"].header
                assert "FOO" in hdulist["PRIMARY"].header
                assert "BAZ" not in hdulist["SCI"].header
            elif only == "SCI":
                assert "TELESCOP" not in hdulist["PRIMARY"].header
                assert "CRVAL1" in hdulist["SCI"].header
                assert "FOO" not in hdulist["PRIMARY"].header
                assert "BAZ" in hdulist["SCI"].header
            else:
                assert "TELESCOP" in hdulist["PRIMARY"].header
                assert "CRVAL1" in hdulist["SCI"].header
                assert "FOO" in hdulist["PRIMARY"].header
                assert "BAZ" in hdulist["SCI"].header

        else:
            assert "FOO" not in hdulist["PRIMARY"].header
            assert "BAZ" not in hdulist["SCI"].header

            if only == "PRIMARY":
                assert "TELESCOP" in hdulist["PRIMARY"].header
                assert "CRVAL1" not in hdulist["SCI"].header
            elif only == "SCI":
                assert "TELESCOP" not in hdulist["PRIMARY"].header
                assert "CRVAL1" in hdulist["SCI"].header
            else:
                assert "TELESCOP" in hdulist["PRIMARY"].header
                assert "CRVAL1" in hdulist["SCI"].header


def test_update_from_dict(tmp_path):
    """Test update method from a dictionary"""
    path = tmp_path / "update.fits"
    with ImageModel((5, 5)) as im:
        update_dict = dict(meta=dict(telescope="JWST", wcsinfo=dict(crval1=5)))
        im.update(update_dict)
        im.save(path)

    with fits.open(path) as hdulist:
        assert "TELESCOP" in hdulist[0].header
        assert "CRVAL1" in hdulist[1].header


def test_mask_model():
    with MaskModel((10, 10)) as dm:
        assert dm.dq.dtype == np.uint32


def test_multislit_model():
    data = np.arange(24).reshape((6, 4))
    err = np.arange(24).reshape((6, 4)) + 2
    wav = np.arange(24).reshape((6, 4)) + 3
    dq = np.arange(24).reshape((6, 4)) + 1
    s0 = SlitDataModel(data=data, err=err, dq=dq, wavelength=wav)
    s1 = SlitDataModel(data=data + 1, err=err + 1, dq=dq + 1, wavelength=wav + 1)

    ms = MultiSlitModel()
    ms.slits.append(s0)
    ms.slits.append(s1)
    ms.meta.instrument.name = 'NIRSPEC'
    ms.meta.exposure.type = 'NRS_IMAGE'
    slit1 = ms[1]
    assert isinstance(slit1, SlitModel)
    assert slit1.meta.instrument.name == 'NIRSPEC'
    assert slit1.meta.exposure.type == 'NRS_IMAGE'
    assert_allclose(slit1.data, data + 1)


def test_slit_from_image():
    data = np.arange(24, dtype=np.float32).reshape((6, 4))
    im = ImageModel(data=data, err=data / 2, dq=data)
    im.meta.instrument.name = "MIRI"
    slit_dm = SlitDataModel(im)
    assert_allclose(im.data, slit_dm.data)
    assert hasattr(slit_dm, 'wavelength')
    # this should be enabled after gwcs starts using non-coordinate inputs
    # assert not hasattr(slit_dm, 'meta')

    slit = SlitModel(im)
    assert_allclose(im.data, slit.data)
    assert_allclose(im.err, slit.err)
    assert hasattr(slit, 'wavelength')
    assert slit.meta.instrument.name == "MIRI"

    im = ImageModel(slit)
    assert type(im) == ImageModel

    im = ImageModel(slit_dm)
    assert type(im) == ImageModel


def test_ifuimage():
    data = np.arange(24).reshape((6, 4))
    im = ImageModel(data=data, err=data / 2, dq=data)
    ifuimage = IFUImageModel(im)
    assert_array_equal(im.data, ifuimage.data)
    assert_array_equal(im.err, ifuimage.err)
    assert_array_equal(im.dq, ifuimage.dq)

    im = ImageModel(ifuimage)
    assert type(im) == ImageModel


def test_abvega_offset_model():
    path = os.path.join(ROOT_DIR, 'nircam_abvega_offset.asdf')
    with ABVegaOffsetModel(path) as model:
        assert isinstance(model, ABVegaOffsetModel)
        assert hasattr(model, 'abvega_offset')
        assert isinstance(model.abvega_offset, Table)
        assert model.abvega_offset.colnames == ['filter', 'pupil', 'abvega_offset']
        model.validate()


@pytest.mark.parametrize("model", [v for v in defined_models.values()])
def test_all_datamodels_init(model):
    """
    Test that all current datamodels can be initialized.
    """
    if model in (DrizProductModel, MultiProductModel, MIRIRampModel):
        with pytest.warns(DeprecationWarning):
            model()
    else:
        model()


def test_meta_date_management(tmp_path):
    model = JwstDataModel({"meta": {"date": "2000-01-01T00:00:00.000"}})
    assert model.meta.date == "2000-01-01T00:00:00.000"
    model.save(str(tmp_path / "test.fits"))
    assert abs((Time.now() - Time(model.meta.date)).value) < 1.0

    model = JwstDataModel()
    assert abs((Time.now() - Time(model.meta.date)).value) < 1.0


def test_ramp_model_zero_frame_open_file(tmpdir):
    """
    Ensures opening a FITS with ZEROFRAME results in a good ZEROFRAME.
    """
    nints, ngroups, nrows, ncols = 2, 10, 5, 5
    dims = nints, ngroups, nrows, ncols
    zdims = (nints, nrows, ncols)

    dbase = 1000.
    zbase = dbase * .75

    data = np.ones(dims, dtype=float) * dbase
    err = np.zeros(dims, dtype=float)
    pdq = np.zeros(dims[-2:], dtype=np.uint32)
    gdq = np.zeros(dims, dtype=np.uint8)

    # Test default exposure zero_frame
    ramp = datamodels.RampModel(data=data, err=err, pixeldq=pdq, groupdq=gdq)

    ramp.meta.exposure.zero_frame = True
    ramp.zeroframe = np.ones(zdims, dtype=ramp.data.dtype) * zbase

    ofile = "my_temp_ramp.fits"
    fname = os.path.join(tmpdir, ofile)
    ramp.save(fname)

    # Check opening a file doesn't change the dimensions
    with datamodels.RampModel(fname) as ramp1:
        assert ramp1.zeroframe.shape == ramp.zeroframe.shape
        zframe0 = ramp.zeroframe
        zframe1 = ramp1.zeroframe
        np.testing.assert_allclose(zframe0, zframe1, 1.e-5)


def test_ramp_model_zero_frame_by_dimensions():
    """
    Ensures creating a RampModel by dimensions results in the correct
    dimensions for ZEROFRAME.
    """
    nints, ngroups, nrows, ncols = 2, 10, 5, 5
    dims = (nints, ngroups, nrows, ncols)
    zdims = (nints, nrows, ncols)

    with datamodels.RampModel(dims) as ramp:
        assert ramp.zeroframe.shape == zdims


@pytest.fixture
def oifits_ami_model():
    m = datamodels.AmiOIModel()
    m.meta.telescope = 'JWST'
    m.meta.origin = 'STScI'
    m.meta.instrument.name = 'NIRISS'
    m.meta.program.pi_name = 'UNKNOWN'
    m.meta.target.proposer_name = 'AB DOR'
    m.meta.observation.date = '2022-06-05'
    m.meta.oifits.array_name = 'g7s6'
    m.meta.oifits.instrument_mode = 'NRM'

    m.array = [
        (
            'A1', 'A1', 1, 0.,
            [ 2.64000000e+00, -1.61653377e-16,  0.00000000e+00], 5.04539835,
            'RADIUS', [ 2.60973996, -0.39856915]
        ),
        (
            'A2', 'A2', 2, 0.,
            [ 1.39996111e-16,  2.28631000e+00,  0.00000000e+00], 5.04539835,
            'RADIUS', [-0.34517145, -2.260104  ]
        ),
        (
            'A3', 'A3', 3, 0.,
            [ 1.32000010e+00, -2.28631000e+00,  0.00000000e+00], 5.04539835,
            'RADIUS', [ 1.65004153,  2.06081941]
        )
    ]
    m.target = [(
        1, 'AB DOR', 82.18704833, -65.44869139, 2000., 0., 0., 0., 'UNKNOWN',
        'OPTICAL', 29.15, 164.421, 0., 0., 65.3199, 0., 'K0V')]
    m.t3 = [
        (
            1, 0., 59735., 0.3772, 0.71125417, 0.73752607, -1.81575411, 0.73752607,
            1.86153485, 2.9549114, -4.32092341, -1.99521297, [1, 2, 3],  0
        ),
    ]
    m.vis = [
        (
            1, 0., 59735., 0.3772, 0.84231787, 0.00669874, -10.57082496, 1.89098664,
            1.86153485,  2.9549114 , [1, 2],  0
        ),
        (
            1, 0., 59735., 0.3772, 0.91448467, 0.00621287, -22.26334374, 1.27637574,
            -2.45938856,  0.95969843, [1, 3],  0
        )
    ]
    m.vis2 = [
        (1, 0., 59735., 0.3772, 0.70949939, 0.01131277,  1.86153485, 2.9549114 , [1, 2],  0),
        (1, 0., 59735., 0.3772, 0.8362822 , 0.0113618 , -2.45938856, 0.95969843, [1, 3],  0)
    ]
    m.wavelength = [(4.817e-06, 2.98e-07)]
    return m


def test_amioi_model_oifits_compliance(tmp_path, oifits_ami_model):
    """
    This test cannot fully test oifits compliance but the schema for the model
    provides some checks. This test checks that the model in the oifits_ami_model
    fixture provides a 'passable' OIFITS file.
    """
    fn = tmp_path / "test.fits"
    oifits_ami_model.save(fn)


@pytest.mark.parametrize('attr',
                         [
                             'meta.telescope',
                             'meta.origin',
                             'meta.instrument.name',
                             'meta.program.pi_name',
                             'meta.target.proposer_name',
                             'meta.observation.date',
                             'meta.oifits.array_name',
                             'meta.oifits.instrument_mode',
                             'array',
                             'target',
                             'wavelength',
                         ])
def test_amioi_model_oifits_keyword_validation(tmp_path, oifits_ami_model, attr):
    """
    Remove some critical keywords or tables and make sure the model fails to
    save/validate This test cannot fully test oifits compliance but at least
    does some sanity checks that a file produced using AmiOiModel produces a
    file containing items that should make it oifits compliant
    """
    fn = tmp_path / "test.fits"

    node = oifits_ami_model
    *branches, leaf = attr.split('.')
    for branch in branches:
        node = getattr(node, branch)
    delattr(node, leaf)

    with pytest.raises(ValidationError):
        oifits_ami_model.save(fn)


@pytest.mark.parametrize('keep', ['vis', 'vis2', 't3'])
def test_amioi_model_oifits_datatable(tmp_path, oifits_ami_model, keep):
    fn = tmp_path / "test.fits"
    for table in ('vis', 'vis2', 't3'):
        if table == keep:
            continue
        delattr(oifits_ami_model, table)

    # since we have 1 data table, this should pass
    oifits_ami_model.save(fn)

    # removing the data table should cause saving to fail
    delattr(oifits_ami_model, keep)
    with pytest.raises(ValidationError):
        oifits_ami_model.save(fn)


def test_dq_def_roundtrip(tmp_path):
    """
    Open a MaskModel with a defined DQ array and dq_def that modifies the
    DQ array on open.  Save the model to a new name.  Open the new model
    to verify the DQ array is equal to the original DQ array.  This validates
    that the dynamic_mask is invertible and computes the same values.
    """
    bname = "jwst_nircam_mask_ref.fits"
    nbname = bname.replace("ref", "ref_dummy")

    fname = os.path.join(ROOT_DIR, bname)
    new_fname = os.path.join(tmp_path, nbname)

    diff = None
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # Open the MaskModel and save it to another file.  It should not save
        # the "uncompressed" DQ array that is in memory, but "recompress" the
        # "uncompressed" DQ array to saved.
        with MaskModel(fname) as mask:
            mask.save(new_fname)

        # Open the original and new mask model to ensure the roundtrip
        # results in the same values.  The ensures the dynamic_mask function
        # that uncompresses and compresses the DQ array based on the dq_def
        # table roundtrips to the same values.
        with MaskModel(fname) as mask, MaskModel(new_fname) as new_mask:
            diff = np.zeros(mask.dq.shape, dtype=int)
            diff[mask.dq != new_mask.dq] = 1

    assert diff is not None
    total_diff = sum(sum(diff))
    assert total_diff == 0


def test_deprecation_data_model():
    """ Tests that inheriting from stdatamodels.jwst.datamodels.DataModel
    raises a DeprecationWarning."""

    with pytest.deprecated_call():
        class Dummy(DataModel):
            pass



from asdf.tags.core import HistoryEntry

def _iterable(values):
    if isinstance(values, str) or not hasattr(values, '__iter__'):
        values = (values,)
    return values

class HistoryList:
    """
    A list that coerces a new value into a HistoryEntry.
    Only a subset of the list interface is implemented.
    """
    def __init__(self, asdf):
        self._context = asdf
        if len(self._context.get_history_entries()):
            self._entries = self._context.get_history_entries()
        else:
            self._context.add_history_entry("fake entry")
            self._entries = self._context.get_history_entries()
            self._entries.clear()

    def __len__(self):
        return len(self._entries)

    def __getitem__(self, key):
        return self._entries[key]

    def __setitem__(self, key, value):
        self.append(value)
        value = self._entries.pop()
        self._entries[key] = value

    def __delitem__(self, key):
        del self._entries[key]

    def __iter__(self):
        return iter(self._entries)

    def __repr__(self):
        return repr(self._entries)

    def __str__(self):
        return str(self._entries)

    def __eq__(self, other):
        if isinstance(other, HistoryList):
            other = other._entries
        else:
            other = _iterable(other)

        if len(self) != len(other):
            return False

        for self_entry, other_entry in zip(self._entries, other):
            if isinstance(other_entry, str):
                if self_entry.get('description') != other_entry:
                    return False
            elif isinstance(other_entry, dict):
                for key in other_entry.keys():
                    if self_entry.get(key) != other_entry.get(key):
                        return False
        return True

    def append(self, value):
        if isinstance(value, HistoryEntry):
            self._entries.append(value)
        else:
            self._context.add_history_entry(value)

    def clear(self):
        self._entries.clear()

    def extend(self, values):
        values = _iterable(values)
        for value in values:
            self.append(value)



from .model_base import JwstDataModel


__all__ = ['CombinedSpecModel']


class CombinedSpecModel(JwstDataModel):
    """
    A data model for combined 1D spectra.

    Parameters
    __________
    spec_table : numpy table
         Combined, extracted spectral data table
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/combinedspec.schema"



from .model_base import JwstDataModel


__all__ = ['RampFitOutputModel']


class RampFitOutputModel(JwstDataModel):
    """
    A data model for the optional output of the ramp fitting step.

    In the parameter definitions below, `n_int` is the number of
    integrations, `max_seg` is the maximum number of segments that
    were fit, `nreads` is the number of reads in an integration, and
    `ny` and `nx` are the height and width of the image.

    Parameters
    __________

    slope : numpy float32 array (n_int, max_seg, ny, nx)
        Segment-specific slope

    sigslope : numpy float32 array (n_int, max_seg, ny, nx)
        Sigma for segment-specific slope

    var_poisson : numpy float32 array (n_int, max_seg, ny, nx)
        Variance due to poisson noise for segment-specific slope

    var_rnoise : numpy float32 array (n_int, max_seg, ny, nx)
        Variance due to read noise for segment-specific slope

    yint : numpy float32 array (n_int, max_seg, ny, nx)
        Segment-specific y-intercept

    sigyint : numpy float32 array (n_int, max_seg, ny, nx)
        Sigma for segment-specific y-intercept

    pedestal : numpy float32 array (n_int, max_seg, ny, nx)
        Pedestal array

    weights : numpy float32 array (n_int, max_seg, ny, nx)
        Weights for segment-specific fits

    crmag : numpy float32 array (n_int, max_seg, ny, nx)
        Approximate CR magnitudes
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/rampfitoutput.schema"



from .reference import ReferenceFileModel

__all__ = ['MirMrsXArtCorrModel']


class MirMrsXArtCorrModel(ReferenceFileModel):
    """
    A data model for MIRI MRS cross-artifact corrections file.

    Parameters
    __________
    init : any
        Any of the initializers supported by `~jwst.datamodels.DataModel`.

    ch1a_table : numpy table
         Cross artifact correction parameters for Channel 1A

    ch1b_table : numpy table
         Cross artifact correction parameters for Channel 1B

    ch1c_table : numpy table
         Cross artifact correction parameters for Channel 1C

    ch2a_table : numpy table
         Cross artifact correction parameters for Channel 2A

    ch2b_table : numpy table
         Cross artifact correction parameters for Channel 2B

    ch2c_table : numpy table
         Cross artifact correction parameters for Channel 2C

    ch3a_table : numpy table
         Cross artifact correction parameters for Channel 3A

    ch3b_table : numpy table
         Cross artifact correction parameters for Channel 3B

    ch3c_table : numpy table
         Cross artifact correction parameters for Channel 3C

    ch4a_table : numpy table
         Cross artifact correction parameters for Channel 4A

    ch4b_table : numpy table
         Cross artifact correction parameters for Channel 4B

    ch4c_table : numpy table
         Cross artifact correction parameters for Channel 4C

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/miri_mrsxartcorr.schema"



from asdf.extension import ManifestExtension

from .converters.jwst_models import (Gwa2SlitConverter, Slit2MsaConverter, LogicalConverter,
                                     NirissSOSSConverter, RefractionIndexConverter,
                                     MIRI_AB2SliceConverter, NIRCAMGrismDispersionConverter,
                                     NIRISSGrismDispersionConverter, GratingEquationConverter,
                                     SnellConverter, Rotation3DToGWAConverter,
                                     CoordsConverter, V23ToSkyConverter)


JWST_TRANSFORM_CONVERTERS = [
    CoordsConverter(),
    Gwa2SlitConverter(),
    Slit2MsaConverter(),
    LogicalConverter(),
    NirissSOSSConverter(),
    RefractionIndexConverter(),
    Rotation3DToGWAConverter(),
    MIRI_AB2SliceConverter(),
    NIRCAMGrismDispersionConverter(),
    NIRISSGrismDispersionConverter(),
    GratingEquationConverter(),
    SnellConverter(),
    V23ToSkyConverter(),
]

# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
TRANSFORM_MANIFEST_URIS = [
    "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-1.0.0",
    "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-0.7.0"
]


TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        uri,
        legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
        converters=JWST_TRANSFORM_CONVERTERS,
    )
    for uri in TRANSFORM_MANIFEST_URIS
]



from .model_base import JwstDataModel


__all__ = ['Level1bModel']


class Level1bModel(JwstDataModel):
    """
    A data model for raw 4D ramps level-1b products.

    Parameters
    __________
    data : numpy uint16 array
         The science data

    zeroframe : numpy uint16 array
         Zeroframe array

    refout : numpy uint16 array
         Reference Output

    group : numpy table
         group parameters table

    int_times : numpy table
         table of times for each integration

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/level1b.schema"



import datetime

import numpy as np
from astropy.io import fits
from astropy.time import Time

from asdf.tags.core import HistoryEntry

from stdatamodels import DataModel


def test_historylist_methods():
    m = DataModel()
    h1 = m.history

    info = "First entry"
    h1.append(info)
    assert h1 == info, "Append new history entry"

    h2 = m.history
    assert h2 == info, "Two history lists point to the same object"

    assert len(h1) == 1, "Length of a history list"

    entry = h1[0]
    assert entry["description"] == info, "Get history list item"

    info += " for real"
    h1[0] = info
    assert h1 == info, "Set history list item"

    del h1[0]
    assert len(h1) == 0, "Delete history list item"

    info = ("First entry", "Second_entry", "Third entry")
    h1.extend(info)
    assert len(h1) == 3, "Length of extended history list"
    assert h1 == info, "Contents of extended history list"

    for entry, item in zip(h1, info):
        assert entry["description"]  == item, "Iterate over history list"

    h1.clear()
    assert len(h1) == 0, "Clear history list"


def test_history_from_model_to_fits(tmpdir):
    tmpfits = str(tmpdir.join('tmp.fits'))
    m = DataModel()
    m.history = [HistoryEntry({
        'description': 'First entry',
        'time': Time(datetime.datetime.now().isoformat())})]
    m.history.append(HistoryEntry({
        'description': 'Second entry',
        'time': Time(datetime.datetime.now().isoformat())
    }))
    m.save(tmpfits)

    with fits.open(tmpfits, memmap=False) as hdulist:
        assert list(hdulist[0].header['HISTORY']) == ["First entry",
                                                      "Second entry"]

    with DataModel(tmpfits) as m2:
        m2 = DataModel()
        m2.update(m)
        m2.history = m.history

        assert m2.history == [{'description': "First entry"},
                              {'description': "Second entry"}]

        m2.save(tmpfits)

    with fits.open(tmpfits, memmap=False) as hdulist:
        assert list(hdulist[0].header['HISTORY']) == ["First entry",
                                                      "Second entry"]


def test_history_from_fits(tmpdir):
    tmpfits = str(tmpdir.join('tmp.fits'))
    header = fits.Header()
    header['HISTORY'] = "First entry"
    header['HISTORY'] = "Second entry"
    fits.writeto(tmpfits, np.array([]), header, overwrite=True)

    with DataModel(tmpfits) as m:
        assert m.history == [{'description': 'First entry'},
                             {'description': 'Second entry'}]

        del m.history[0]
        m.history.append(HistoryEntry({'description': "Third entry"}))
        assert m.history == [{'description': "Second entry"},
                             {'description': "Third entry"}]
        m.save(tmpfits)

    with DataModel(tmpfits) as m:
        assert m.history == [{'description': "Second entry"},
                             {'description': "Third entry"}]



from .model_base import JwstDataModel


__all__ = ['CubeModel']


class CubeModel(JwstDataModel):
    """
    A data model for 3D image cubes.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array

    zeroframe : numpy float32 array
         Zero frame array

    area : numpy float32 array
         Pixel area map array

    int_times : numpy table
         table of times for each integration

    wavelength : numpy float32 array
         Wavelength array

    var_poisson : numpy float32 array
         Integration-specific variances of slope due to Poisson noise

    var_rnoise : numpy float32 array
         Integration-specific variances of slope due to read noise
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/cube.schema"

    def __init__(self, init=None, **kwargs):

        super(CubeModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



from asdf import schema as mschema
from astropy.io import fits
from numpy.testing import assert_array_equal
import numpy as np
import pytest

from stdatamodels.jwst import datamodels
from stdatamodels.jwst.datamodels import ImageModel, JwstDataModel, RampModel


@pytest.fixture
def fits_file(tmp_path):
    path = str(tmp_path / "miri_ramp.fits")
    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    data = np.zeros((5, 35, 40, 32))
    image = fits.ImageHDU(data=data, name="SCI", ver=1)
    hdulist.append(image)
    header = hdulist[0].header
    # header["DATAMODL"] = "JwstDataModel"
    # Add invalid keyword
    header["INSTRUME"] = "MIRI"
    hdulist.writeto(path)
    return path


def test_from_fits(fits_file):
    with RampModel(fits_file) as dm:
        assert dm.meta.instrument.name == 'MIRI'
        assert dm.shape == (5, 35, 40, 32)


def test_delete(fits_file):
    with JwstDataModel() as dm:
        dm.meta.instrument.name = 'NIRCAM'
        assert dm.meta.instrument.name == 'NIRCAM'
        del dm.meta.instrument.name
        assert dm.meta.instrument.name is None


def test_fits_without_sci():
    schema = {
        "allOf": [
            mschema.load_schema("http://stsci.edu/schemas/jwst_datamodel/core.schema",
                                resolve_references=True),
            {
                "type": "object",
                "properties": {
                    "coeffs": {
                        'max_ndim': 1,
                        'fits_hdu': 'COEFFS',
                        'datatype': 'float32'
                    }
                }
            }
        ]
    }

    hdulist = fits.HDUList()
    hdulist.append(fits.PrimaryHDU())
    hdulist.append(fits.ImageHDU(name='COEFFS', data=np.array([0.0], np.float32)))

    with JwstDataModel(hdulist, schema=schema) as dm:
        assert_array_equal(dm.coeffs, [0.0])


def test_casting():
    with RampModel((5, 5, 10, 10)) as dm:
        total = dm.data.sum()
        dm.data = dm.data + 2
        assert dm.data.sum() > total


def test_resave_duplication_bug(tmp_path):
    """
    An issue in asdf (https://github.com/asdf-format/asdf/issues/1232)
    resulted in duplication of data when a model was read from and then
    written to a fits file. A resave results in array data being written to
    both hdus and as internal blocks within the asdf tree (which is then
    stored in the asdf hdu).
    """
    fn1 = tmp_path / "test1.fits"
    fn2 = tmp_path / "test2.fits"

    arr = np.zeros((1000, 100), dtype='f4')
    m = ImageModel(arr)
    m.save(fn1)

    with datamodels.open(fn1) as m2:
        m2.save(fn2)

    with fits.open(fn1) as ff1, fits.open(fn2) as ff2:
        assert ff1['ASDF'].size == ff2['ASDF'].size



from .reference import ReferenceFileModel


__all__ = ['PsfMaskModel']


class PsfMaskModel(ReferenceFileModel):
    """
    A data model for coronagraphic 2D PSF mask reference files

    Parameters
    __________
    data : numpy float32 array
         The PSF mask
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/psfmask.schema"



from stdatamodels.dynamicdq import dynamic_mask
from .dqflags import pixel
from .reference import ReferenceFileModel


__all__ = ['NirspecFlatModel', 'NirspecQuadFlatModel']


class NirspecFlatModel(ReferenceFileModel):
    """A data model for NIRSpec flat-field reference files.

    Parameters
    __________
    data : numpy float32 array
         NIRSpec flat-field reference data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error estimate

    wavelength : numpy table
         Table of wavelengths for image planes

    flat_table : numpy table
         Table for quickly varying component of flat field

    dq_def : numpy table
         DQ flag definitions
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_flat.schema"

    def __init__(self, init=None, **kwargs):
        super(NirspecFlatModel, self).__init__(init=init, **kwargs)

        if self.dq is not None or self.dq_def is not None:
            self.dq = dynamic_mask(self, pixel)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err


class NirspecQuadFlatModel(ReferenceFileModel):
    """A data model for NIRSpec flat-field files that differ by quadrant.

    Parameters
    __________
    quadrants.items.data : numpy float32 array


    quadrants.items.dq : numpy uint32 array


    quadrants.items.err : numpy float32 array


    quadrants.items.wavelength : numpy table
         Table of wavelengths for image planes

    quadrants.items.flat_table : numpy table
         Table for quickly varying component of flat field

    dq_def : numpy table
         DQ flag definitions
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nirspec_quad_flat.schema"

    def __init__(self, init=None, **kwargs):
        if isinstance(init, NirspecFlatModel):
            super(NirspecQuadFlatModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.quadrants.append(self.quadrants.item())
            self.quadrants[0].data = init.data
            self.quadrants[0].dq = init.dq
            self.quadrants[0].err = init.err
            self.quadrants[0].wavelength = init.wavelength
            self.quadrants[0].flat_table = init.flat_table
            self.quadrants[0].dq_def = init.dq_def
            self.quadrants[0].dq = dynamic_mask(self.quadrants[0])
            return

        super(NirspecQuadFlatModel, self).__init__(init=init, **kwargs)



from .reference import ReferenceFileModel


__all__ = ['ReadnoiseModel']


class ReadnoiseModel(ReferenceFileModel):
    """
    A data model for 2D readnoise.

    Parameters
    __________
    data : numpy float32 array
         Read noise
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/readnoise.schema"



from .model_base import JwstDataModel


__all__ = ['SegmentationMapModel']


class SegmentationMapModel(JwstDataModel):
    """
    A data model for 2D segmentation maps

    Parameters
    __________
    data : numpy uint32 array
         The segmentation map
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/segmap.schema"



"""
Test datamodel.open
"""

import contextlib
import os
import os.path
from pathlib import Path, PurePath
import warnings

import pytest
import numpy as np
from astropy.io import fits
from stdatamodels import DataModel
from stdatamodels.validate import ValidationError, ValidationWarning

from stdatamodels.jwst.datamodels import (JwstDataModel, ImageModel,
                                          RampModel, CubeModel, ReferenceFileModel, ReferenceImageModel,
                                          ReferenceCubeModel, ReferenceQuadModel)
from stdatamodels.jwst import datamodels
from stdatamodels.jwst.datamodels import util

import asdf

# Define artificial memory size
MEMORY = 100  # 100 bytes


@pytest.mark.parametrize('guess', [True, False])
def test_guess(guess):
    """Test the guess parameter to the open func"""
    path = Path(t_path('test.fits'))

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "model_type not found")
        if guess is None or guess:
            # Default is to guess at the model type.
            with datamodels.open(path, guess=guess) as model:
                assert isinstance(model, JwstDataModel)
        else:
            # Without guessing, the call should fail.
            with pytest.raises(TypeError):
                with datamodels.open(path, guess=guess) as model:
                    pass


def test_mirirampmodel_deprecation(tmp_path):
    """Test that a deprecated MIRIRampModel can be opened"""
    path = str(tmp_path / "ramp.fits")
    # Create a MIRIRampModel, working around the deprecation.
    model = datamodels.RampModel((1, 1, 10, 10))
    model.save(path)
    hduls = fits.open(path, mode='update')
    hduls[0].header['datamodl'] = 'MIRIRampModel'
    hduls.close()

    # Test it.
    with pytest.warns(DeprecationWarning):
        miri_ramp = datamodels.open(path)
    assert isinstance(miri_ramp, datamodels.RampModel)


@pytest.fixture
def mock_get_available_memory(monkeypatch):
    def mock(include_swap=True):
        available = MEMORY
        if include_swap:
            available *= 2
        return available
    monkeypatch.setattr(util, 'get_available_memory', mock)


@pytest.mark.parametrize(
    'allowed_env, allowed_explicit, result',
    [
        (None, None, True),  # Perform no check.
        (0.1, None, False),  # Force too little memory.
        (0.1, 1.0, True),    # Explicit overrides environment.
        (1.0, 0.1, False),   # Explicit overrides environment.
        (None, 0.1, False),  # Explicit overrides environment.
    ]
)
def test_check_memory_allocation_env(monkeypatch, mock_get_available_memory,
                                     allowed_env, allowed_explicit, result):
    """Check environmental control over memory check"""
    if allowed_env is None:
        monkeypatch.delenv('DMODEL_ALLOWED_MEMORY', raising=False)
    else:
        monkeypatch.setenv('DMODEL_ALLOWED_MEMORY', str(allowed_env))

    # Allocate amount that would fit at 100% + swap.
    can_allocate, required = util.check_memory_allocation(
        (MEMORY // 2, 1), allowed=allowed_explicit,
    )
    assert can_allocate is result


@pytest.mark.parametrize(
    'dim, allowed, include_swap, result',
    [
        (MEMORY // 2, 1.0, True, True),    # Fit within memory and swap
        (MEMORY // 2, 1.0, False, False),  # Does not fit without swap
        (MEMORY, 1.0, True, False),        # Does not fit at all
        (MEMORY, None, True, True),        # Check disabled
        (MEMORY // 2, 0.1, True, False),   # Does not fit in restricted memory
    ]
)
def test_check_memory_allocation(mock_get_available_memory, dim, allowed, include_swap, result):
    """Check general operation of check_memory_allocation"""
    can_allocate, required = util.check_memory_allocation(
        (dim, 1), allowed=allowed, include_swap=include_swap
    )
    assert can_allocate is result


def test_open_from_pathlib():
    """Test opening a PurePath object"""
    path = Path(t_path('test.fits'))
    assert isinstance(path, PurePath)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "model_type not found")
        with datamodels.open(path) as model:
            assert isinstance(model, JwstDataModel)


def test_open_fits():
    """Test opening a model from a FITS file"""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "model_type not found")
        fits_file = t_path('test.fits')
        with datamodels.open(fits_file) as model:
            assert isinstance(model, JwstDataModel)


def test_open_none():
    with datamodels.open() as model:
        assert isinstance(model, JwstDataModel)


def test_open_shape():
    shape = (50, 20)
    with datamodels.open(shape) as model:
        assert isinstance(model, ImageModel)
        assert model.shape == shape


def test_open_illegal():
    with pytest.raises(ValueError):
        init = 5
        datamodels.open(init)


def test_open_hdulist(tmp_path):
    hdulist = fits.HDUList()
    primary = fits.PrimaryHDU()
    data = np.empty((50, 50), dtype=np.float32)
    science = fits.ImageHDU(data=data, name='SCI', ver=1)
    hdulist.append(primary)
    hdulist.append(science)

    # datamodels.open() can't open pathlib objects
    path = str(tmp_path / "jwst_image.fits")
    hdulist.writeto(path)

    with datamodels.open(hdulist) as model:
        assert isinstance(model, ImageModel)

    with pytest.warns(datamodels.util.NoTypeWarning) as record:
        with datamodels.open(path) as model:
            assert isinstance(model, ImageModel)
            assert len(record) == 1
            assert "model_type not found" in record[0].message.args[0]


def test_open_ramp(tmp_path):
    """Open 4D data without a DQ as RampModel"""
    path = str(tmp_path / "ramp.fits")
    shape = (2, 3, 4, 5)
    with fits.HDUList(fits.PrimaryHDU()) as hdulist:
        hdulist.append(fits.ImageHDU(data=np.zeros(shape), name="SCI", ver=1))
        hdulist.writeto(path)

    with pytest.warns(datamodels.util.NoTypeWarning):
        with datamodels.open(path) as model:
            assert isinstance(model, RampModel)


def test_open_cube(tmp_path):
    """Open 3D data as CubeModel"""
    path = str(tmp_path / "ramp.fits")
    shape = (2, 3, 4)
    with fits.HDUList(fits.PrimaryHDU()) as hdulist:
        hdulist.append(fits.ImageHDU(data=np.zeros(shape), name="SCI", ver=1))
        hdulist.writeto(path)

    with pytest.warns(datamodels.util.NoTypeWarning):
        with datamodels.open(path) as model:
            assert isinstance(model, CubeModel)


@pytest.mark.parametrize("model_class, shape", [
    (ReferenceFileModel, None),
    (ReferenceImageModel, (10, 10)),
    (ReferenceCubeModel, (3, 3, 3)),
    (ReferenceQuadModel, (2, 2, 2, 2)),
])
def test_open_reffiles(tmp_path, model_class, shape):
    """Try opening files with a REFTYPE keyword and different data/dq shapes"""
    path = str(tmp_path / "reffile.fits")
    with fits.HDUList(fits.PrimaryHDU()) as hdulist:
        hdulist["PRIMARY"].header.append(("REFTYPE", "foo"))
        if shape is not None:
            hdulist.append(fits.ImageHDU(data=np.zeros(shape), name="SCI", ver=1))
            hdulist.append(fits.ImageHDU(data=np.zeros(shape, dtype=np.uint), name="DQ", ver=1))
        hdulist.writeto(path)

    with pytest.warns(datamodels.util.NoTypeWarning):
        with datamodels.open(path) as model:
            assert isinstance(model, model_class)


@pytest.mark.parametrize("suffix", [".asdf", ".fits"])
def test_open_readonly(tmp_path, suffix):
    """Test opening a FITS-format datamodel that is read-only on disk"""
    path = str(tmp_path / f"readonly{suffix}")

    with ImageModel(data=np.zeros((10, 10))) as model:
        model.meta.telescope = 'JWST'
        model.meta.instrument.name = 'NIRCAM'
        model.meta.instrument.detector = 'NRCA4'
        model.meta.instrument.channel = 'SHORT'
        model.save(path)

    os.chmod(path, 0o440)
    assert os.access(path, os.W_OK) is False

    with datamodels.open(path) as model:
        assert model.meta.telescope == 'JWST'
        assert isinstance(model, ImageModel)


# Utilities
def t_path(partial_path):
    """Construction the full path for test files"""
    test_dir = os.path.join(os.path.dirname(__file__), 'data')
    return os.path.join(test_dir, partial_path)


@pytest.mark.parametrize("suffix", ["asdf", "fits"])
def test_open_asdf_no_datamodel_class(tmp_path, suffix):
    path = str(tmp_path / f"no_model.{suffix}")
    model = DataModel()
    model.save(path)

    # Note: only the fits open emits a "model_type not found" warning.  Both
    # fits and asdf should behave the same
    ctx = pytest.warns(util.NoTypeWarning) if suffix == 'fits' else contextlib.nullcontext()
    with ctx:
        with datamodels.open(path) as m:
            assert isinstance(m, DataModel)


def test_open_asdf(tmp_path):
    path = str(tmp_path / "straight_asdf.asdf")
    tree = {"foo": 42, "bar": 13, "seq": np.arange(100)}
    with asdf.AsdfFile(tree) as af:
        af.write_to(path)

    with pytest.warns(util.NoTypeWarning):
        with datamodels.open(path) as m:
            assert isinstance(m, DataModel)


def test_open_kwargs_asdf(tmp_path):
    """
    Test that unrecognized kwargs to the datamodels.open function
    are passed on to the model class constructor.
    """
    file_path = tmp_path / "test.asdf"

    with pytest.warns(ValidationWarning):
        model = ImageModel((4, 4), pass_invalid_values=True)
        model.meta.instrument.name = "CELESTRON"
        model.save(file_path)

    with pytest.raises(ValidationError):
        with datamodels.open(file_path, strict_validation=True) as model:
            model.validate()


def test_open_kwargs_fits(tmp_path):
    """
    Test that unrecognized kwargs to the datamodels.open function
    are passed on to the model class constructor.  Similar to the
    above, except the invalid file must be created differently
    because DataModel can't save an invalid .fits file.
    """
    file_path = tmp_path / "test.fits"

    model = ImageModel((4, 4))
    model.save(file_path)

    with fits.open(file_path, mode="update") as hdul:
        hdul[0].header["INSTRUME"] = "CELESTRON"
        hdul.flush()

    with pytest.raises(ValidationError):
        with datamodels.open(file_path, strict_validation=True) as model:
            model.validate()



from .reference import ReferenceFileModel


__all__ = ['TrapParsModel']


class TrapParsModel(ReferenceFileModel):
    """
    A data model for trap capture and decay parameters.

    Parameters
    __________
    trappars_table : numpy table
         Trap capture and decay parameters
         A table with three columns for trap-capture parameters and one
         column for the trap-decay parameter.  Each row of the table is
         for a different trap family.
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/trappars.schema"



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
import warnings
import numpy as np
from collections.abc import Mapping
from astropy.io import fits

from asdf.tags.core import ndarray

from . import util
from . import validate
from . import schema as mschema

import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


__all__ = ['ObjectNode', 'ListNode']


def _is_struct_array(val):
    return (isinstance(val, (np.ndarray, fits.FITS_rec)) and
            val.dtype.names is not None and val.dtype.fields is not None)


def _is_struct_array_precursor(val):
    return isinstance(val, list) and isinstance(val[0], tuple)


def _is_struct_array_schema(schema):
    return (isinstance(schema['datatype'], list) and
            any('name' in t for t in schema['datatype']))


def _cast(val, schema):
    val = _unmake_node(val)
    if val is None:
        return None

    if 'datatype' in schema:
        # Handle lazy array
        if isinstance(val, ndarray.NDArrayType):
            val = val._make_array()

        if (_is_struct_array_schema(schema) and len(val) and
            (_is_struct_array_precursor(val) or _is_struct_array(val))):
            # we are dealing with a structured array. Because we may
            # modify schema (to add shape), we make a deep copy of the
            # schema here:
            schema = copy.deepcopy(schema)

            for t, v in zip(schema['datatype'], val[0]):
                if not isinstance(t, Mapping):
                    continue

                aval = np.asanyarray(v)
                shape = aval.shape
                val_ndim = len(shape)

                # make sure that if 'ndim' is specified for a field,
                # it matches the dimensionality of val's field:
                if 'ndim' in t and val_ndim != t['ndim']:
                    raise ValueError(
                        "Array has wrong number of dimensions. "
                        "Expected {}, got {}".format(t['ndim'], val_ndim)
                    )

                if 'max_ndim' in t and val_ndim > t['max_ndim']:
                    raise ValueError(
                        "Array has wrong number of dimensions. "
                        "Expected <= {}, got {}".format(t['max_ndim'], val_ndim)
                    )

                # if shape of a field's value is not specified in the schema,
                # add it to the schema based on the shape of the actual data:
                if 'shape' not in t:
                    t['shape'] = shape

        dtype = ndarray.asdf_datatype_to_numpy_dtype(schema['datatype'])
        val = util.gentle_asarray(val, dtype)

        if dtype.fields is not None:
            val = _as_fitsrec(val)

    if 'ndim' in schema and len(val.shape) != schema['ndim']:
        raise ValueError(
            "Array has wrong number of dimensions.  Expected {}, got {}"
            .format(schema['ndim'], len(val.shape)))

    if 'max_ndim' in schema and len(val.shape) > schema['max_ndim']:
        raise ValueError(
            "Array has wrong number of dimensions.  Expected <= {}, got {}"
            .format(schema['max_ndim'], len(val.shape)))

    if isinstance(val, np.generic) and np.isscalar(val):
        val = val.item()

    return val


def _as_fitsrec(val):
    """
    Convert a numpy record into a fits record if it is not one already
    """
    if isinstance(val, fits.FITS_rec):
        return val
    else:
        coldefs = fits.ColDefs(val)
        uint = any(c._pseudo_unsigned_ints for c in coldefs)
        fits_rec = fits.FITS_rec(val)
        fits_rec._coldefs = coldefs
        # FITS_rec needs to know if it should be operating in pseudo-unsigned-ints mode,
        # otherwise it won't properly convert integer columns with TZEROn before saving.
        fits_rec._uint = uint
        return fits_rec


def _get_schema_type(schema):
    """
    Create a list of types used by a schema and its subschemas when
    the subschemas are joined by combiners. Then return a type string
    if all the types are the same or 'mixed' if they differ
    """
    def callback(subschema, path, combiner, types, recurse):
        if 'type' in subschema:
            types.append(subschema['type'])

        has_combiner = ('anyOf' in subschema.keys() or
                        'allOf' in subschema.keys())
        return not has_combiner

    types = []
    mschema.walk_schema(schema, callback, types)

    schema_type = None
    for a_type in types:
        if schema_type is None:
            schema_type = a_type
        elif schema_type != a_type:
            schema_type = 'mixed'
            break

    return schema_type


def _make_default_array(attr, schema, ctx):
    dtype = schema.get('datatype')
    if dtype is not None:
        dtype = ndarray.asdf_datatype_to_numpy_dtype(dtype)
    ndim = schema.get('ndim', schema.get('max_ndim'))
    default = schema.get('default', None)
    primary_array_name = ctx.get_primary_array_name()

    if attr == primary_array_name:
        if ctx.shape is not None:
            shape = ctx.shape
        elif ndim is not None:
            shape = tuple([0] * ndim)
        else:
            shape = (0,)
    else:
        if dtype.names is not None:
            if ndim is None:
                shape = (0,)
            else:
                shape = tuple([0] * ndim)
            default = None
        else:
            has_primary_array_shape = False
            if primary_array_name is not None:
                primary_array = getattr(ctx, primary_array_name, None)
                has_primary_array_shape = primary_array is not None

            if has_primary_array_shape:
                if ndim is None:
                    shape = primary_array.shape
                else:
                    if attr == "zeroframe":
                        dims = primary_array.shape
                        if len(dims) != 4:
                            error_msg = "To allocate ZEROFRAME, the primary array must "
                            error_msg += f"have 4 dimensions, but has {len(dims)}."
                            raise IndexError(error_msg)
                        shape = (dims[0], dims[2], dims[3])
                    else:
                        shape = primary_array.shape[-ndim:]
            elif ndim is None:
                shape = (0,)
            else:
                shape = tuple([0] * ndim)

    array = np.empty(shape, dtype=dtype)
    if default is not None:
        array[...] = default
    return array


def _make_default(attr, schema, ctx):
    if 'max_ndim' in schema or 'ndim' in schema or 'datatype' in schema:
        return _make_default_array(attr, schema, ctx)
    elif 'default' in schema:
        return schema['default']
    else:
        schema_type = _get_schema_type(schema)
        if schema_type == 'object':
            return {}
        elif schema_type == 'array':
            return []
        else:
            return None


def _make_node(attr, instance, schema, ctx, parent):
    if isinstance(instance, dict):
        return ObjectNode(attr, instance, schema, ctx, parent)
    elif isinstance(instance, list):
        return ListNode(attr, instance, schema, ctx, parent)
    else:
        return instance


def _unmake_node(obj):
    if isinstance(obj, Node):
        return obj.instance
    return obj


def _get_schema_for_property(schema, attr):
    subschema = schema.get('properties', {}).get(attr, None)
    if subschema is not None:
        return subschema
    for combiner in ['allOf', 'anyOf']:
        for subschema in schema.get(combiner, []):
            subsubschema = _get_schema_for_property(subschema, attr)
            if subsubschema != {}:
                return subsubschema
    return {}


def _get_schema_for_index(schema, i):
    items = schema.get('items', {})
    if isinstance(items, list):
        if i >= len(items):
            return {}
        else:
            return items[i]
    else:
        return items

def _find_property(schema, attr):
    subschema = _get_schema_for_property(schema, attr)
    if subschema == {}:
        find = False
    else:
        find = 'default' in subschema
    return find

class Node():
    def __init__(self, attr, instance, schema, ctx, parent):
        self._name = attr
        self._instance = instance
        self._schema = schema
        self._ctx = ctx
        self._parent = parent

    def _validate(self):
        return validate.value_change(self._name, self._instance, self._schema, self._ctx)

    @property
    def instance(self):
        return self._instance

class ObjectNode(Node):
    def __dir__(self):
        added = set(self._schema.get('properties', {}).keys())
        return sorted(set(super().__dir__()) | added)

    def __eq__(self, other):
        if isinstance(other, ObjectNode):
            return self._instance == other._instance
        else:
            return self._instance == other

    def _is_deprecated(self, attr):
        return "deprecated_properties" in self._schema and attr in self._schema["deprecated_properties"]

    def _warn_deprecated_attr(self, attr):
        node = self
        path = []
        while node._parent is not None:
            if not isinstance(node, ListNode):
                path.append(node._name)
            node = node._parent

        path.reverse()

        old_attr = ".".join(path + [attr])

        property_path = self._schema["deprecated_properties"][attr]
        parts = property_path.split("/")
        for part in parts:
            if part == "..":
                path.pop()
            else:
                path.append(part)

        new_attr = ".".join(path)

        warnings.warn(f"Attribute '{old_attr}' has been deprecated.  Use '{new_attr}' instead.", DeprecationWarning)

    def _get_deprecated(self, attr):
        property_path = self._schema["deprecated_properties"][attr]
        node = self
        parts = property_path.split("/")
        for part in parts[:-1]:
            if part == "..":
                node = node._parent
            else:
                node = getattr(node, part)
        return node, parts[-1]

    def __getattr__(self, attr):

        if attr.startswith('_'):
            raise AttributeError('No attribute {0}'.format(attr))

        if self._is_deprecated(attr):
            self._warn_deprecated_attr(attr)
            node, new_attr = self._get_deprecated(attr)
            return getattr(node, new_attr)

        schema = _get_schema_for_property(self._schema, attr)
        try:
            val = self._instance[attr]
        except KeyError:
            if schema == {}:
                raise AttributeError("No attribute '{0}'".format(attr))

            val = _make_default(attr, schema, self._ctx)
            if val is not None:
                self._instance[attr] = val

        if isinstance(val, dict):
            node = ObjectNode(attr, val, schema, self._ctx, self)
        elif isinstance(val, list):
            node = ListNode(attr, val, schema, self._ctx, self)
        else:
            node = val

        return node

    def __setattr__(self, attr, val):
        if attr.startswith('_'):
            self.__dict__[attr] = val
        elif self._is_deprecated(attr):
            self._warn_deprecated_attr(attr)
            node, new_attr = self._get_deprecated(attr)
            setattr(node, new_attr, val)
        else:
            schema = _get_schema_for_property(self._schema, attr)
            if val is None:
                val = _make_default(attr, schema, self._ctx)
            val = _cast(val, schema)

            node = ObjectNode(attr, val, schema, self._ctx, self)
            if self._ctx._validate_on_assignment:
                if node._validate():
                    self._instance[attr] = val
            else:
                self._instance[attr] = val

    def __delattr__(self, attr):
        if attr.startswith('_'):
            del self.__dict__[attr]
        elif self._is_deprecated(attr):
            self._warn_deprecated_attr(attr)
            node, new_attr = self._get_deprecated(attr)
            delattr(node, new_attr)
        else:
            schema = _get_schema_for_property(self._schema, attr)
            if validate.value_change(attr, None, schema, self._ctx) or self._ctx._pass_invalid_values:
                try:
                    del self._instance[attr]
                except KeyError:
                    raise AttributeError(
                        "Attribute '{0}' missing".format(attr))

    def __iter__(self):
        return NodeIterator(self)

    def hasattr(self, attr):
        return attr in self._instance

    def items(self):
        # Return a (key, value) tuple for the node
        for key in self:
            val = self
            for field in key.split('.'):
                val = getattr(val, field)
            yield (key, val)

class ListNode(Node):
    def __cast(self, other):
        if isinstance(other, ListNode):
            return other._instance
        return other

    def __repr__(self):
        return repr(self._instance)

    def __eq__(self, other):
        return self._instance == self.__cast(other)

    def __ne__(self, other):
        return self._instance != self.__cast(other)

    def __contains__(self, item):
        return item in self._instance

    def __len__(self):
        return len(self._instance)

    def __getitem__(self, i):
        schema = _get_schema_for_index(self._schema, i)
        return _make_node(self._name, self._instance[i], schema, self._ctx, self)

    def __setitem__(self, i, val):
        schema = _get_schema_for_index(self._schema, i)
        val =  _cast(val, schema)
        node = ObjectNode(self._name, val, schema, self._ctx, self)
        if self._ctx._validate_on_assignment:
            if node._validate():
                self._instance[i] = val
        else:
            self._instance[i] = val

    def __delitem__(self, i):
        del self._instance[i]
        if self._ctx._validate_on_assignment:
            self._validate()

    def __getslice__(self, i, j):
        if isinstance(self._schema['items'], list):
            r = range(*(slice(i, j).indices(len(self._instance))))
            schema_parts = [
                _get_schema_for_index(self._schema, x) for x in r
            ]
        else:
            schema_parts = self._schema['items']
        schema = {'type': 'array', 'items': schema_parts}
        return _make_node(self._name, self._instance[i:j], schema, self._ctx, self)

    def __setslice__(self, i, j, other):
        parts = _unmake_node(other)
        parts = [_cast(x, _get_schema_for_index(self._schema, k))
                 for (k, x) in enumerate(parts)]
        self._instance[i:j] = _unmake_node(other)
        if self._ctx._validate_on_assignment:
            self._validate()

    def __delslice__(self, i, j):
        del self._instance[i:j]
        if self._ctx._validate_on_assignment:
            self._validate()

    def append(self, item):
        schema = _get_schema_for_index(self._schema, len(self._instance))
        item = _cast(item, schema)
        node = ObjectNode(self._name, item, schema, self._ctx, self)
        if self._ctx._validate_on_assignment:
            if node._validate():
                self._instance.append(item)
        else:
            self._instance.append(item)


    def insert(self, i, item):
        schema = _get_schema_for_index(self._schema, i)
        item = _cast(item, schema)
        node = ObjectNode(self._name, item, schema, self._ctx, self)
        if self._ctx._validate_on_assignment:
            if node._validate():
                self._instance.insert(i, item)
        else:
            self._instance.insert(i, item)

    def pop(self, i=-1):
        schema = _get_schema_for_index(self._schema, 0)
        x = self._instance.pop(i)
        return _make_node(self._name, x, schema, self._ctx, self)

    def remove(self, item):
        self._instance.remove(item)

    def count(self, item):
        return self._instance.count(item)

    def index(self, item):
        return self._instance.index(item)

    def reverse(self):
        self._instance.reverse()

    def sort(self, *args, **kwargs):
        self._instance.sort(*args, **kwargs)

    def extend(self, other):
        for part in _unmake_node(other):
            self.append(part)

    def item(self, **kwargs):
        assert isinstance(self._schema['items'], dict)
        node = ObjectNode(self._name, kwargs, self._schema['items'],
                          self._ctx, self)
        if not self._ctx._validate_on_assignment:
            return node
        if not node._validate():
            node = None
        return node

class NodeIterator:
    """
    An iterator for a node which flattens the hierachical structure
    """
    def __init__(self, node):
        self.key_stack = []
        self.iter_stack = [iter(node._instance.items())]

    def __iter__(self):
        return self

    def __next__(self):
        while self.iter_stack:
            try:
                key, val = next(self.iter_stack[-1])
            except StopIteration:
                self.iter_stack.pop()
                if self.iter_stack:
                    self.key_stack.pop()
                continue

            if isinstance(val, dict):
                self.key_stack.append(key)
                self.iter_stack.append(iter(val.items()))
            else:
                return '.'.join(self.key_stack + [key])

        raise StopIteration

def put_value(path, value, tree):
    """
    Put a value at the given path into tree, replacing it if it is
    already present.

    Parameters
    ----------
    path : list of str or int
        The path to the element.

    value : any
        The value to place

    tree : JSON object tree
    """
    cursor = tree
    for i in range(len(path) - 1):
        part = path[i]
        if isinstance(part, int):
            while len(cursor) <= part:
                cursor.append({})
            cursor = cursor[part]
        else:
            if isinstance(path[i + 1], int) or path[i + 1] == 'items':
                cursor = cursor.setdefault(part, [])
            else:
                cursor = cursor.setdefault(part, {})

    if isinstance(path[-1], int):
        while len(cursor) <= path[-1]:
            cursor.append({})
    cursor[path[-1]] = value


def merge_tree(a, b):
    """
    Merge elements from tree `b` into tree `a`.
    """
    def recurse(a, b):
        if isinstance(b, dict):
            if not isinstance(a, dict):
                return copy.deepcopy(b)
            for key, val in b.items():
                a[key] = recurse(a.get(key), val)
            return a
        return copy.deepcopy(b)

    recurse(a, b)
    return a



from astropy import time
from datetime import datetime
import jsonschema
import pytest

from stdatamodels.jwst.datamodels import JwstDataModel


def test_strict_validation_enum():
    with JwstDataModel(strict_validation=True) as dm:
        assert dm.meta.instrument.name is None
        with pytest.raises(jsonschema.ValidationError):
            # FOO is not in the allowed enumerated values
            dm.meta.instrument.name = 'FOO'


def test_strict_validation_type():
    with JwstDataModel(strict_validation=True) as dm:
        with pytest.raises(jsonschema.ValidationError):
            # Schema requires a float
            dm.meta.target.ra = "FOO"


def test_strict_validation_date():
    with JwstDataModel(strict_validation=True) as dm:
        time_obj = time.Time(dm.meta.date)
        assert isinstance(time_obj, time.Time)
        date_obj = datetime.strptime(dm.meta.date, '%Y-%m-%dT%H:%M:%S.%f')
        assert isinstance(date_obj, datetime)



import re


__all__ = ["multiple_replace"]


def multiple_replace(string, rep_dict):
    """Single-pass replacement of multiple substrings

    Similar to `str.replace`, except that a dictionary of replacements
    can be specified.

    The replacements are done in a single-pass. This means that a previous
    replacement will not be replaced by a subsequent match.

    Parameters
    ----------
    string: str
        The source string to have replacements done on it.

    rep_dict: dict
        The replacements were key is the input substring and
        value is the replacement

    Returns
    -------
    replaced: str
        New string with the replacements done

    Examples
    --------
    Basic example that also demonstrates the single-pass nature.
    If the replacements where chained, the result would have been
    'lamb lamb'

    >>> multiple_replace('button mutton', {'but': 'mut', 'mutton': 'lamb'})
    'mutton lamb'

    """
    pattern = re.compile(
        "|".join([re.escape(k) for k in sorted(rep_dict, key=len, reverse=True)]),
        flags=re.DOTALL
    )
    return pattern.sub(lambda x: rep_dict[x.group(0)], string)



from .model_base import JwstDataModel


__all__ = ['TrapsFilledModel']


class TrapsFilledModel(JwstDataModel):
    """
    A data model for the number of traps filled for a detector, for
    persistence.

    Parameters
    __________
    data : numpy float32 array
        Traps filled
        The map of the number of traps filled over the detector, with
        one plane for each "trap family."
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/trapsfilled.schema"



from .model_base import JwstDataModel


__all__ = ['QuadModel']


class QuadModel(JwstDataModel):
    """
    A data model for 4D image arrays.

    Parameters
    __________
    data : numpy float32 array
         The science data

    dq : numpy uint32 array
         Data quality array

    err : numpy float32 array
         Error array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/quad.schema"

    def __init__(self, init=None, **kwargs):
        super(QuadModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



"""
Implementation
--------------

The flags are implemented as "bit flags": Each flag is assigned a bit position
in a byte, or multi-byte word, of memory. If that bit is set, the flag assigned
to that bit is interpreted as being set or active.

The data structure that stores bit flags is just the standard Python `int`,
which provides 32 bits. Bits of an integer are most easily referred to using
the formula `2**bit_number` where `bit_number` is the 0-index bit of interest.
"""
from astropy.nddata.bitmask import interpret_bit_flags as ap_interpret_bit_flags
from stdatamodels.basic_utils import multiple_replace


def interpret_bit_flags(bit_flags, flip_bits=None, mnemonic_map=None):
    """Converts input bit flags to a single integer value (bit mask) or `None`.

    Wraps `astropy.nddate.bitmask.interpret_bit_flags`, allowing the
    bit mnemonics to be used in place of integers.

    Parameters
    ----------
    bit_flags : int, str, list, None
        See `astropy.nddate.bitmask.interpret_bit_flags`.
        Also allows strings using Roman mnemonics

    flip_bits : bool, None
        See `astropy.nddata.bitmask.interpret_bit_flags`.

    mnemonic_map : {str: int[,...]}
        Dictionary associating the mnemonic string to an integer value
        representing the set bit for that mnemonic.

    Returns
    -------
    bitmask : int or None
        Returns an integer bit mask formed from the input bit value or `None`
        if input ``bit_flags`` parameter is `None` or an empty string.
        If input string value was prepended with '~' (or ``flip_bits`` was set
        to `True`), then returned value will have its bits flipped
        (inverse mask).
    """
    if mnemonic_map is None:
        raise TypeError("`mnemonic_map` is a required argument")
    bit_flags_dm = bit_flags
    if isinstance(bit_flags, str):
        dm_flags = {
            key: str(val)
            for key, val in mnemonic_map.items()
        }
        bit_flags_dm = multiple_replace(bit_flags, dm_flags)

    return ap_interpret_bit_flags(bit_flags_dm, flip_bits=flip_bits)


def dqflags_to_mnemonics(dqflags, mnemonic_map):
    """Interpret value as bit flags and return the mnemonics

    Parameters
    ----------
    dqflags : int-like
        The value to interpret as DQ flags

    mnemonic_map: {str: int[,...]}
        Dictionary associating the mnemonic string to an integer value
        representing the set bit for that mnemonic.

    Returns
    -------
    mnemonics : {str[,...]}
        Set of mnemonics represented by the set bit flags

    Examples
    --------
    >>> pixel = {'GOOD':             0,      # No bits set, all is good
    ...          'DO_NOT_USE':       2**0,   # Bad pixel. Do not use
    ...          'SATURATED':        2**1,   # Pixel saturated during exposure
    ...          'JUMP_DET':         2**2,   # Jump detected during exposure
    ...          }

    >>> group = {'GOOD':       pixel['GOOD'],
    ...          'DO_NOT_USE': pixel['DO_NOT_USE'],
    ...          'SATURATED':  pixel['SATURATED'],
    ...          }

    >>> dqflags_to_mnemonics(1, pixel)
    {'DO_NOT_USE'}

    >>> dqflags_to_mnemonics(7, pixel)             #doctest: +SKIP
    {'JUMP_DET', 'DO_NOT_USE', 'SATURATED'}

    >>> dqflags_to_mnemonics(7, pixel) == {'JUMP_DET', 'DO_NOT_USE', 'SATURATED'}
    True

    >>> dqflags_to_mnemonics(1, mnemonic_map=pixel)
    {'DO_NOT_USE'}

    >>> dqflags_to_mnemonics(1, mnemonic_map=group)
    {'DO_NOT_USE'}
    """
    mnemonics = {
        mnemonic
        for mnemonic, value in mnemonic_map.items()
        if (dqflags & value)
    }
    return mnemonics



import asdf
from astropy.io import fits

from . import fits_support


__all__ = [
    'write',
    'open'
]


def write(filename, tree, hdulist=None, **kwargs):
    """Write ASDF data inside a fits file

    Parameters
    ----------
    filename : str or path
        Filename where the resulting fits file containing the ASDF
        data will be saved. This is passed on to
        :func:`astropy.io.fits.HDUList.writeto`

    tree : ASDF tree or dict
        ASDF data to save in the fits file

    kwargs : variable keyword arguments
        Passed on to :func:`astropy.io.fits.HDUList.writeto`
    """
    hdulist = fits_support.to_fits(tree, None, hdulist=hdulist)  # no custom schema
    hdulist.writeto(filename, **kwargs)


def open(filename_or_hdu, **kwargs):
    """Read ASDF data embedded in a fits file

    Parameters
    ----------
    filename_or_hdu : str, path, `astropy.io.fits.HDUList`
        Filename of the fits file or an open `astropy.io.fits.HDUList`
        containing the ASDF data. If a filename is provided it
        will be opened with :func:`astropy.io.fits.open`.

    kwargs : variable keyword arguments
        Passed on to :func:`asdf.open`

    Returns
    -------
    af : :obj:`asdf.AsdfFile`
        :obj:`asdf.AsdfFile` created from ASDF data embeded in the opened
        fits file.
    """

    is_hdu = isinstance(filename_or_hdu, fits.HDUList)
    hdulist = filename_or_hdu if is_hdu else fits.open(filename_or_hdu)
    if 'ignore_missing_extensions' not in kwargs:
        kwargs['ignore_missing_extensions'] = False
    af = fits_support.from_fits_asdf(hdulist, **kwargs)

    if is_hdu:
        # no need to wrap close if input was an HDUList
        return af

    # on close, close hdulist
    def wrap_close(af, hdulist):
        def close():
            asdf.AsdfFile.close(af)
            hdulist.close()
        return close

    af.close = wrap_close(af, hdulist)
    return af



from .model_base import JwstDataModel


__all__ = ['GuiderRawModel', 'GuiderCalModel']


class GuiderRawModel(JwstDataModel):
    """
    A data model for Guide Star pipeline raw data files

    Parameters
    __________
    data : numpy float32 array
         The science data

    err : numpy float32 array
         Error array

    dq : numpy uint32 array
         Data quality array

    planned_star_table : numpy table
         Planned reference star table

    flight_star_table : numpy table
         Flight reference star table

    pointing_table : numpy table
         Pointing table

    centroid_table : numpy table
         Centroid packet table

    track_sub_table : numpy table
         Track subarray data table
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/guider_raw.schema"

    def __init__(self, init=None, **kwargs):

        super(GuiderRawModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err


class GuiderCalModel(JwstDataModel):
    """
    A data model for Guide Star pipeline calibrated files

    Parameters
    __________
    data : numpy float32 array
         The science data

    err : numpy float32 array
         Error array

    dq : numpy uint32 array
         Data quality array

    planned_star_table : numpy table
         Planned reference star table

    flight_star_table : numpy table
         Flight reference star table

    pointing_table : numpy table
         Pointing table

    centroid_table : numpy table
         Centroid packet table

    track_sub_table : numpy table
         Track subarray data table
    """

    schema_url = "http://stsci.edu/schemas/jwst_datamodel/guider_cal.schema"

    def __init__(self, init=None, **kwargs):

        super(GuiderCalModel, self).__init__(init=init, **kwargs)

        # Implicitly create arrays
        self.dq = self.dq
        self.err = self.err



import warnings

import pytest
import asdf
from jsonschema import ValidationError
import numpy as np

from stdatamodels.validate import ValidationWarning
from models import BasicModel, ValidationModel, RequiredModel


class _DoesNotRaiseContext:
    """
    Dummy context manager for use in parametrized tests, for non-raising cases.
    """
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __repr__(self):
        return "does_not_raise"

# Define a singleton to use
does_not_raise = _DoesNotRaiseContext()


def test_scalar_attribute_assignment():
    model = ValidationModel()

    assert model.meta.string_attribute is None
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.string_attribute = "foo"
    assert model.meta.string_attribute == "foo"

    with pytest.warns(ValidationWarning):
        model.meta.string_attribute = 42
    assert model.meta.string_attribute == "foo"

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.string_attribute = None
    assert model.meta.string_attribute is None


def test_object_attribute_assignment():
    model = ValidationModel()

    assert model.meta.object_attribute.string_attribute is None
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.object_attribute = {"string_attribute": "foo"}
    assert model.meta.object_attribute.string_attribute == "foo"

    with pytest.warns(ValidationWarning):
        model.meta.object_attribute = "bar"
    assert model.meta.object_attribute.string_attribute == "foo"

    with pytest.warns(ValidationWarning):
        model.meta.object_attribute = {"string_attribute": 42}
    assert model.meta.object_attribute.string_attribute == "foo"

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.object_attribute = None
    assert model.meta.object_attribute.string_attribute is None


def test_list_attribute_ssignment():
    model = ValidationModel()

    assert len(model.meta.list_attribute) == 0
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.list_attribute.append({"string_attribute": "foo"})
    assert model.meta.list_attribute[0].string_attribute == "foo"

    with pytest.warns(ValidationWarning):
        model.meta.list_attribute.append({"string_attribute": 42})
    assert len(model.meta.list_attribute) == 1
    assert model.meta.list_attribute[0].string_attribute == "foo"

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.list_attribute = None
    assert len(model.meta.list_attribute) == 0


def test_object_assignment_with_nested_null():
    model = ValidationModel()

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.object_attribute = {"string_attribute": None}


@pytest.mark.xfail(reason="validation of a required attribute not yet implemented", strict=True)
def test_required_attribute_assignment():
    model = RequiredModel()

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.required_attribute = "foo"

    with pytest.warns(ValidationWarning):
        model.meta.required_attribute = None


@pytest.mark.xfail(reason="validation of required attributes not yet implemented", strict=True)
def test_validation_on_delete():
    model = RequiredModel()

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.required_keyword = "foo"

    with pytest.warns(ValidationWarning):
        del model.meta.required_keyword
    assert model.meta.required_keyword == "foo"

    model = RequiredModel(pass_invalid_values=True)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.required_keyword = "foo"

    with pytest.warns(ValidationWarning):
        del model.meta.required_keyword
    assert model.meta.required_keyword is None


@pytest.mark.parametrize(
    "init_value, env_value, passed",
    [
        (None, None, False),
        (True, None, True),
        (False, None, False),
        (None, "true", True),
        (None, "false", False),
    ],
)
def test_pass_invalid_values_attribute_assignment(monkeypatch, init_value,
                                                  env_value, passed):
    if env_value is not None:
        monkeypatch.setenv("PASS_INVALID_VALUES", env_value)

    model = ValidationModel(pass_invalid_values=init_value)

    with pytest.warns(ValidationWarning):
        model.meta.string_attribute = 42

    if passed:
        assert model.meta.string_attribute == 42
    else:
        assert model.meta.string_attribute is None


@pytest.mark.parametrize("suffix",
    [
        "asdf",
        pytest.param("fits",
            marks=pytest.mark.xfail(reason="save to FITS raises error, not just warning", strict=True))
    ])
def test_pass_invalid_values_on_write(tmp_path, suffix):
    file_path = tmp_path / f"test.{suffix}"
    model = ValidationModel(pass_invalid_values=True)
    with pytest.warns(ValidationWarning):
        model.meta.string_attribute = 42
    with pytest.warns(ValidationWarning):
        model.save(file_path)

    with asdf.open(file_path) as af:
        assert af["meta"]["string_attribute"] == 42


@pytest.mark.parametrize(
    "init_value, env_value, expected_context_manager",
    [
        (None, None, pytest.warns(ValidationWarning)),
        (True, None, pytest.raises(ValidationError)),
        (False, None, pytest.warns(ValidationWarning)),
        (None, "true", pytest.raises(ValidationError)),
        (None, "false", pytest.warns(ValidationWarning)),
    ],
)
def test_strict_validation_attribute_assignment(monkeypatch, init_value, env_value,
                                                expected_context_manager):
    if env_value is not None:
        monkeypatch.setenv("STRICT_VALIDATION", env_value)

    model = ValidationModel(strict_validation=init_value)

    with expected_context_manager:
        model.meta.string_attribute = 42
    assert model.meta.string_attribute is None


def test_validate():
    model = ValidationModel(pass_invalid_values=True)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        model.meta.string_attribute = "foo"
        model.validate()

    with pytest.warns(ValidationWarning):
        model.meta.string_attribute = 42
    assert model.meta.string_attribute == 42

    with pytest.warns(ValidationWarning):
        model.validate()


@pytest.mark.xfail(reason="validation on init not yet implemented for ASDF files", strict=True)
def test_validation_on_init(tmp_path):
    with asdf.AsdfFile() as af:
        af["meta"] = {"string_attribute": "foo"}

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            ValidationModel(af)

        af["meta"]["string_attribute"] = 42
        with pytest.warns(ValidationWarning):
            ValidationModel(af)


@pytest.mark.parametrize("suffix",
    [
        "asdf",
        pytest.param("fits",
            marks=pytest.mark.xfail(reason="save to FITS raises error, not just warning", strict=True))
    ])
def test_validation_on_write(tmp_path, suffix):
    file_path = tmp_path / f"test.{suffix}"
    model = ValidationModel(pass_invalid_values=True)
    with pytest.warns(ValidationWarning):
        model.meta.string_attribute = 42

    with pytest.warns(ValidationWarning):
        model.save(file_path)


@pytest.mark.parametrize(
    "init_value, env_value, expected_context_manager, string_attribute_value",
    [
        (None, None, pytest.warns(ValidationWarning), None),
        (True, None, pytest.warns(ValidationWarning), None),
        (False, None, does_not_raise, 42),
        (None, "true", pytest.warns(ValidationWarning), None),
        (None, "false", does_not_raise, 42),
    ],
)
def test_validate_on_assignment(monkeypatch, init_value, env_value,
                                expected_context_manager, string_attribute_value):
    if env_value is not None:
        monkeypatch.setenv("VALIDATE_ON_ASSIGNMENT", env_value)
    model = ValidationModel(validate_on_assignment=init_value)

    with expected_context_manager:
        model.meta.string_attribute = 42  # Bad assignment
    assert model.meta.string_attribute is string_attribute_value


@pytest.mark.parametrize(
    "init_value, warning_class, string_attribute_value",
    [
        (True, ValidationWarning, "bar"),
        (False, None, 42),
    ],
)
def test_validate_on_assignment_setitem(init_value, warning_class,
                                        string_attribute_value):
    model = ValidationModel(validate_on_assignment=init_value)

    # Check values assigned that are valid
    value = "foo"
    model.meta.list_attribute.append({"string_attribute": value})
    assert model.meta.list_attribute[0].string_attribute == value

    value2 = "bar"
    model.meta.list_attribute[0] = {"string_attribute": value2}
    assert model.meta.list_attribute[0].string_attribute == value2

    # Now check invalid assignments.  Currently string_attribute="bar".  Try
    # assigning an invalid type
    value3 = 42
    if warning_class is None:
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            model.meta.list_attribute[0] = {"string_attribute": value3}
    else:
        with pytest.warns(warning_class):
            model.meta.list_attribute[0] = {"string_attribute": value3}

    assert model.meta.list_attribute[0].string_attribute == string_attribute_value


@pytest.mark.parametrize(
    "validate_on_assignment, warning_class, string_attribute_value",
    [
        (True, ValidationWarning, "bar"),
        (False, None, 42),
    ],
)
def test_validate_on_assignment_insert(validate_on_assignment, warning_class,
                                       string_attribute_value):
    model = ValidationModel(validate_on_assignment=validate_on_assignment)

    model.meta.list_attribute.insert(0, {"string_attribute": "bar"})
    assert model.meta.list_attribute[0].string_attribute == "bar"

    if warning_class is None:
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            model.meta.list_attribute.insert(0, {"string_attribute": 42})
    else:
        with pytest.warns(warning_class):
            model.meta.list_attribute.insert(0, {"string_attribute": 42})
    assert model.meta.list_attribute[0].string_attribute == string_attribute_value


# --------------------------------------------------------------------
# Validation flag interaction testing.

# This test is the same as one above, but with strict_validation=True.
# This ensures strict_validation=True doesn't effect this test.

@pytest.mark.parametrize(
    "validate_on_assignment, strict_validation, expected_context_manager, value",
    [
        (True, False, pytest.warns(ValidationWarning), None),
        (True, True, pytest.raises(ValidationError), None),
        (False, False, does_not_raise, 42),
        (False, True, does_not_raise, 42),
    ],
)
def test_validate_on_assignment_strict_validation(tmp_path, validate_on_assignment,
                                                  strict_validation,
                                                  expected_context_manager, value):
    model = ValidationModel(validate_on_assignment=validate_on_assignment,
                            strict_validation=strict_validation)

    with expected_context_manager:
        model.meta.string_attribute = 42
    assert model.meta.string_attribute is value


@pytest.mark.parametrize(
    "validate_on_assignment, pass_invalid_values, expected_context_manager, value",
    [
        (True, False, pytest.warns(ValidationWarning), None),
        (True, True, pytest.warns(ValidationWarning), 42),
        (False, False, does_not_raise, 42),
        (False, True, does_not_raise, 42),
    ],
)
def test_validate_on_assignment_pass_invalid_values(validate_on_assignment,
                                                    pass_invalid_values,
                                                    expected_context_manager,
                                                    value):
    model = ValidationModel(validate_on_assignment=validate_on_assignment,
                            pass_invalid_values=pass_invalid_values)

    # pass_invalid_values=True allows for assignment,
    # even with validate_on_assignment=True
    with expected_context_manager:
        model.meta.string_attribute = 42  # Bad assignment
    assert model.meta.string_attribute == value


def test_ndarray_validation(tmp_path):
    file_path = tmp_path / "test.asdf"

    # Wrong dtype
    with asdf.AsdfFile() as af:
        af["data"] = np.ones((4, 4), dtype=np.float64)
        af.write_to(file_path)

    with pytest.raises(ValidationError, match="Array datatype 'float64' is not compatible with 'float32'"):
        with BasicModel(file_path, strict_validation=True, validate_arrays=True) as model:
            model.validate()

    # Wrong dimensions
    with asdf.AsdfFile() as af:
        af["data"] = np.ones((4,), dtype=np.float32)
        af.write_to(file_path)

    with pytest.raises(ValidationError, match="Wrong number of dimensions: Expected 2, got 1"):
        with BasicModel(file_path, strict_validation=True, validate_arrays=True) as model:
            model.validate()



from .reference import ReferenceFileModel


__all__ = ['SpecTraceModel', 'SpecTraceSingleModel']


class SpecTraceModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS spectral trace reference files.

    This model has a special member `trace` that can be used to
    deal with an entire spectral trace at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import SpecTraceSingleModel
       >>> spectrace_model = SpecTraceModel()
       >>> spectrace_model.trace.append(SpecTraceSingleModel())
       >>> spectrace_model.trace[0] # doctest: +SKIP
       <SpecTraceSingleModel>

    If `init` is a `SpecTraceSingleModel` instance, an empty `SpecTraceSingleModel`
    will be created and assigned to attribute `trace[0]`, and the `data`
    attribute from the input `SpecTraceSingleModel` instance will be copied to
    the first element of `trace`.  `SpecTraceSingleModel` objects can be appended
    to the `trace` attribute by using its `append` method.

    Parameters
    __________
    trace.items.data : numpy table
         Spectral trace data

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/spectrace.schema"

    def __init__(self, init=None, **kwargs):

        if isinstance(init, SpecTraceSingleModel):
            super(SpecTraceModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.trace.append(self.trace.item())
            self.trace[0].data = init.data

        super(SpecTraceModel, self).__init__(init=init, **kwargs)


class SpecTraceSingleModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS spectral trace data.

    Parameters
    __________
    data : numpy table
         Trace values
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/spectracesingle.schema"



from .reference import ReferenceFileModel


__all__ = ['WaveMapModel', 'WaveMapSingleModel']


class WaveMapModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS wavelength map reference files.

    This model has a special member `map` that can be used to
    deal with an entire wavelength map at a time.  It behaves like a list::

       >>> from stdatamodels.jwst.datamodels import WaveMapSingleModel, WaveMapModel
       >>> wavemap_model = WaveMapModel()
       >>> wavemap_model.map.append(WaveMapSingleModel())
       >>> wavemap_model.map[0] # doctest: +SKIP
       <WaveMapSingleModel>

    If `init` is a `WaveMapSingleModel` instance, an empty `WaveMapSingleModel`
    will be created and assigned to attribute `map[0]`, and the `data`
    attribute from the input `WaveMapSingleModel` instance will be copied to
    the first element of `map`.  `WaveMapSingleModel` objects can be appended
    to the `map` attribute by using its `append` method.

    Parameters
    __________
    map.items.data : numpy data array
         Wavelength map data

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/wavemap.schema"

    def __init__(self, init=None, **kwargs):

        if isinstance(init, WaveMapSingleModel):
            super(WaveMapModel, self).__init__(init=None, **kwargs)
            self.update(init)
            self.map.append(self.map.item())
            self.map[0].data = init.data

        super(WaveMapModel, self).__init__(init=init, **kwargs)


class WaveMapSingleModel(ReferenceFileModel):
    """
    A data model for NIRISS SOSS wavelength map data.

    Parameters
    __________
    data : numpy float32 array
         Wavelength values
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/wavemapsingle.schema"



from stdatamodels import dqflags

PIXEL = {'GOOD':             0,      # No bits set, all is good
         'DO_NOT_USE':       2**0,   # Bad pixel. Do not use.
         'SATURATED':        2**1,   # Pixel saturated during exposure
         'JUMP_DET':         2**2,   # Jump detected during exposure
         'DROPOUT':          2**3,   # Data lost in transmission
         'OUTLIER':          2**4,   # Flagged by outlier detection (was RESERVED_1)
         'PERSISTENCE':      2**5,   # High persistence (was RESERVED_2)
         'AD_FLOOR':         2**6,   # Below A/D floor (0 DN, was RESERVED_3)
         'RESERVED_4':       2**7,   #
         'UNRELIABLE_ERROR': 2**8,   # Uncertainty exceeds quoted error
         'NON_SCIENCE':      2**9,   # Pixel not on science portion of detector
         'DEAD':             2**10,  # Dead pixel
         'HOT':              2**11,  # Hot pixel
         'WARM':             2**12,  # Warm pixel
         'LOW_QE':           2**13,  # Low quantum efficiency
         'RC':               2**14,  # RC pixel
         'TELEGRAPH':        2**15,  # Telegraph pixel
         'NONLINEAR':        2**16,  # Pixel highly nonlinear
         'BAD_REF_PIXEL':    2**17,  # Reference pixel cannot be used
         'NO_FLAT_FIELD':    2**18,  # Flat field cannot be measured
         'NO_GAIN_VALUE':    2**19,  # Gain cannot be measured
         'NO_LIN_CORR':      2**20,  # Linearity correction not available
         'NO_SAT_CHECK':     2**21,  # Saturation check not available
         'UNRELIABLE_BIAS':  2**22,  # Bias variance large
         'UNRELIABLE_DARK':  2**23,  # Dark variance large
         'UNRELIABLE_SLOPE': 2**24,  # Slope variance large (i.e., noisy pixel)
         'UNRELIABLE_FLAT':  2**25,  # Flat variance large
         'OPEN':             2**26,  # Open pixel (counts move to adjacent pixels)
         'ADJ_OPEN':         2**27,  # Adjacent to open pixel
         'UNRELIABLE_RESET': 2**28,  # Sensitive to reset anomaly
         'MSA_FAILED_OPEN':  2**29,  # Pixel sees light from failed-open shutter
         'OTHER_BAD_PIXEL':  2**30,  # A catch-all flag
         'REFERENCE_PIXEL':  2**31,  # Pixel is a reference pixel
         }


def test_dqflags():
    assert dqflags.dqflags_to_mnemonics(1, PIXEL) == {'DO_NOT_USE'}
    assert dqflags.dqflags_to_mnemonics(7, PIXEL) == {'JUMP_DET', 'DO_NOT_USE', 'SATURATED'}
    assert dqflags.interpret_bit_flags('DO_NOT_USE + WARM', mnemonic_map=PIXEL) == 4097
